/*
 *  Programmazione Grafica 3D - Cap. 1
 *  by Andrea Fasano
 *
 *  Esempio di utilizzo della palette, ed implementazione delle funzioni
 *  per il fade in ed il fade out
 *  (versione con sfarfallio)
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 */

#include <stdio.h>

#include <dos.h>
#include <conio.h>
#include <sys/farptr.h>
#include <stdlib.h>

     #include <dpmi.h>          /* definisce __dpmi_segment_to_descriptor */
     short video;

typedef struct {
   char red, green, blue;
} TYPE_PALETTE;

TYPE_PALETTE GetColor (char num_color)
 {
  TYPE_PALETTE palette;

  outp (0x3c7, num_color);
  palette.red   = inp (0x3c9);
  palette.green = inp (0x3c9);
  palette.blue  = inp (0x3c9);

  return palette;
 }

void GetPalette (TYPE_PALETTE *palette)
 {
  int i;

  for (i=0; i<256; i++)
    palette[i] = GetColor (i);
 }

void SetColor (char num_color, char red, char green, char blue)
 {
  outp (0x3c8, num_color);
  outp (0x3c9, red);
  outp (0x3c9, green);
  outp (0x3c9, blue);
 }

void SetPalette (TYPE_PALETTE *palette)
 {
  int i;

  for (i=0; i<256; i++)
    SetColor (i, palette[i].red, palette[i].green, palette[i].blue);
 }

void SetPaletteToBlack (TYPE_PALETTE *palette)
 {
  int i;

  for (i=0; i<256; i++) {
      palette[i].red = 0;
      palette[i].green = 0;
      palette[i].blue = 0;
  }
  SetPalette (palette);
}

void Fade_Out (void)
 {
  int i, j;
  TYPE_PALETTE tmp;

  for (i=0; i<64; i++)
   {
    for (j=0; j<256; j++) {
      tmp = GetColor (j);
      if (tmp.red > 0)   tmp.red--;
      if (tmp.green > 0) tmp.green--;
      if (tmp.blue > 0)  tmp.blue--;
      SetColor (j, tmp.red, tmp.green, tmp.blue);
    }
   }
 }

void Fade_In (TYPE_PALETTE *palette)
 {
  int i, j;
  TYPE_PALETTE tmp;

  for (i=0; i<64; i++)
   {
    for (j=0; j<256; j++) {
      tmp = GetColor (j);
      if (tmp.red < palette[j].red)     tmp.red++;
      if (tmp.green < palette[j].green) tmp.green++;
      if (tmp.blue < palette[j].blue)   tmp.blue++;
      SetColor (j, tmp.red, tmp.green, tmp.blue);
    }
   }
 }

void Set_Text_Mode (void)
 {
  union REGS registers;

  registers.h.ah = 0x00;                /* Servizio 0 dell' interrupt 0x10 */
  registers.h.al = 0x03;                  /* Numero del nuovo modo grafico */
  int386 (0x10, &registers, &registers);
 }

void Create_Selector (void)
{
  video = __dpmi_segment_to_descriptor(0xa000);
  _farsetsel(video);
}

void Set_Vga_Mode (void)
 {
  union REGS registers;

  registers.h.ah = 0x00;		/* Servizio 0 dell' interrupt 0x10 */
  registers.h.al = 0x13;		/* Numero del nuovo modo grafico */
  int386 (0x10, &registers, &registers);
 }

void PutPixel (int x, int y, char color)
 {
  _farnspokeb(x + (y*320), color);
 }


void main (void)
 {
  int x, y, n;
  TYPE_PALETTE screenpal[256], blackpal[256];

  /* Inizializziamo la variabile screenpal con valori casuali */
  screenpal[0].red=0;
  screenpal[0].green=0;
  screenpal[0].blue=0;

  for (n=1; n<256; n++)
    {
      screenpal[n].red = rand () % 64;
      screenpal[n].green = rand () % 64;
      screenpal[n].blue = rand () % 64;
    }

 /* Azzeriamo invece la palette blackpal */
  for (n=0; n<256; n++)
    {
      blackpal[n].red = 0;
      blackpal[n].green = 0;
      blackpal[n].blue = 0;
    }

 /* Attivazione modalit grafica */
  Create_Selector();      /* DJGPP */
  Set_Vga_Mode();

 /* Oscuriamo temporaneamente lo schermo per un
    disegno "nascosto" dei pixel */
  SetPalette (blackpal);

 /* Riempiamo il monito con pixel aventi colori casuali.
     NB: Ancora non si vedono! */
  for (y=0; y<200; y++)
   for (x=0; x<320; x++)
     PutPixel (x, y, rand() % 256);
  getch();

  Fade_In (screenpal);
  getch ();

  Fade_Out ();
  getch ();

  /* Rimettiamo le cose a posto */
  Set_Text_Mode ();
 }
