/*
 *  Programmazione Grafica 3D - Cap. 1
 *  by Andrea Fasano
 *
 *  Esempio di disegno di una linea tramite l'algoritmo di Bresenam
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 */

#include <stdio.h>
     #include <dos.h>
     #include <dpmi.h>          /* definisce __dpmi_segment_to_descriptor */
     short video;

void Create_Selector (void)
{
  video = __dpmi_segment_to_descriptor(0xa000);
  _farsetsel(video);
}

void PutPixel (int x, int y, char color)
 {
  _farnspokeb(x + (y*320), color);
 }

void PutLine (int x1, int y1, int x2, int y2, char color)
{
 int i, deltax, deltay, numpixels,
     d, dinc1, dinc2,
     x, xinc1, xinc2,
     y, yinc1, yinc2;

  /* Calcoliamo il deltax ed il deltay della linea, ovvero il numero di
     pixel presenti a livello orizzontale e verticale. Utilizziamo la
     funzione abs() poich a noi interessa il loro valore assoluto. */
  deltax = abs(x2 - x1);
  deltay = abs(y2 - y1);

  /* Adesso controlliamo se la linea  pi "orizzontale" o "verticale",
     ed inizializziamo in maniera appropriate le variabili di comodo. */
  if (deltax >= deltay)
    {
      /* La linea risulta maggiormente schiacciata sull' ascissa */
      numpixels = deltax + 1;

      /* La nostra variabile decisionale, basata sulla x della linea */	
      d = (2 * deltay) - deltax; 

      /* Settiamo gli incrementi */
      dinc1 = deltay * 2;
      dinc2 = (deltay - deltax) * 2;
      xinc1 = xinc2 = 1;
      yinc1 = 0;
      yinc2 = 1;
    }
  else
    {
      /* La linea risulta maggiormente schiacciata sull' ordinata*/
      numpixels = deltay + 1;

      /* La nostra variabile decisionale, basata sulla y della linea */	 	
      d = (2 * deltax) - deltay;

      /* Settiamo gli incrementi */		
      dinc1 = deltax * 2;
      dinc2 = (deltax - deltay) * 2;
      xinc1 = 0;
      xinc2 = 1;
      yinc1 = yinc2 = 1;
    }

  /* Eseguiamo un controllo per settare il corretto 
      andamento della linea */
  if (x1 > x2)
    {
      xinc1 = - xinc1;
      xinc2 = - xinc2;
    }
  if (y1 > y2)
    {
      yinc1 = - yinc1;
      yinc2 = - yinc2;
    }

  /* Settiamo le coordinate iniziali  */
  x = x1;
  y = y1;

  /* Stampiamo la linea */
  for (i=0; i<numpixels; i++)
    {
      PutPixel(x, y, color);

      /* Scegliamo l' incremento del "passo" a seconda dellla
          variabile decisionale */	
      if (d < 0)
        {
          d = d + dinc1;
          x = x + xinc1;
          y = y + yinc1;
        }
      else
        {
          d = d + dinc2;
          x = x + xinc2;
          y = y + yinc2;
        }
    }
}

/* queste due procedure seguenti hanno subito delle modifiche */

void Put_Horiz_Line (int x1, int x2, int y, char color)
 {
  int len, start, n;

  /* Calcoliamo la lunghezza della linea, ovvero il 
     numero di pixel da stampare */
  len = abs (x2 - x1) + 1;

  if (x1>x2)
      x1=x2;

  /* Stampiamo la linea! */
  start = (x1 + (y*320));
  for (n=0;n<len;n++)
  {
       _farnspokeb(start, color);
       start++;
  }
 }

void Put_Vert_Line (int y1, int y2, int x, char color)
 {
  int h, n, start;

  /* Calcoliamo l' altezza della linea, ovvero il
      numero di pixel da stampare */
  h = abs (y2 - y1) + 1;

  if (y1>y2)
      y1=y2;

  /* Calcoliamo l' indirizzo iniziale del primo punto */
  start = (x + (y1*320));
  for (n=0;n<h;n++)
  {
       _farnspokeb(start, color);
       start+=320;
  }
 }

void Set_Vga_Mode (void)
 {
  union REGS registers;

  registers.h.ah = 0x00;		/* Servizio 0 dell' interrupt 0x10 */
  registers.h.al = 0x13;		/* Numero del nuovo modo grafico */
  int386 (0x10, &registers, &registers);
 }

void Set_Text_Mode (void)
 {
  union REGS registers;

  registers.h.ah = 0x00;                /* Servizio 0 dell' interrupt 0x10 */
  registers.h.al = 0x03;                /* Numero del nuovo modo grafico */
  int386 (0x10, &registers, &registers);
 }


void main (void)
{
 Create_Selector();
 Set_Vga_Mode ();

 /* varie prove */
 Put_Horiz_Line (0, 319, 0, 100);
 Put_Horiz_Line (319, 0, 199, 100);
 Put_Vert_Line (0, 199, 0, 100);
 Put_Vert_Line (199, 0, 319, 100);

 PutLine (0, 0, 319, 199, 60);
 PutLine (319, 0, 0, 199, 60);
 PutLine (80, 0, 240, 199, 60);
 PutLine (80, 199, 240, 0, 60);

 PutLine (1, 1, 80, 50, 10); // oppure (80, 50, 1, 1, 10)
 PutLine (1, 1, 50, 80, 10); // oppure (50, 80, 1, 1, 10)
 PutLine (1, 80, 50, 1, 10); // oppure (50, 1, 1, 80, 10)
 PutLine (1, 50, 80, 1, 10); // oppure (80, 1, 1, 50, 10)

 getch();

 Set_Text_Mode();
}
