/*
 *  Programmazione Grafica 3D - Cap. 1
 *  by Andrea Fasano
 *
 *  Esempio di stampa di un pixel (quarta versione)
 *
 *  Ho effettuato dei test plottando fino a 200 milioni
 *  di punti e, con alcune ottimizzazioni attivate,
 *  si rileva che le versioni 1 e 3 sono praticamente
 *  equivalenti, e cos pure la 2 e la 4, che sono
 *  un po' pi veloci della coppia precedente.
 *  Senza le ottimizzazioni questo divario probabilmente
 *  sar maggiore. Questi sono tutti modi sicuri di
 *  tracciare un punto in DJGGP. Esiste anche un altro
 *  modo, che  pi veloce, per  meno protetto,
 *  perch prevede di disabilitare in modo trucchettoso
 *  la protezione della memoria disponibile nel modo detto
 *  appunto "protetto"; perci usatelo solo nelle versioni
 *  finali dei vostri programmi di demo (dopo che siete
 *  sicuri che il programma  corretto). Lo trovate in
 *  ex_3f.c
 *  Per maggiori informazioni consultate le seguenti sezioni
 *  delle faq di djgpp nella directory djgpp/faq:
 *
 *     10.2 Accessing the video memory
 *
 *     18.6 Fast access to absolute addresses
 *
 *     18.4 How to move data between your program and conventional memory
 *
 *  I metodi 3 e 4 offrono una maggiore protezione: non  permesso di
 *  leggere/scrivere pi di 64Kb con il selettore definito. La memoria
 *  video in 320x200  per 320*200=64000 bytes e non 64Kb, perci
 *   meglio definire un selettore apposito che copra esattamente
 *  64000 bytes. Il file ex_3e.c riporta questo codice che in
 *  pratica, tra i metodi presentati,  quello che ha pi vantaggi
 *  nella fase di debugging e il pi veloce, escludendo quello
 *  presentato in ex_3f.c, se volete garantire la massima
 *  compatibilit con tutti i DPMI server.
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 */

#include <stdio.h>
#include <dos.h>

     #include <dpmi.h>          /* definisce __dpmi_segment_to_descriptor */

     short video;

void Create_Selector (void)
{
  video = __dpmi_segment_to_descriptor(0xa000);
  _farsetsel(video);
}

void Set_Text_Mode (void)
 {
  union REGS registers;

  registers.h.ah = 0x00;                /* Servizio 0 dell' interrupt 0x10 */
  registers.h.al = 0x03;                /* Numero del nuovo modo grafico   */
  int386 (0x10, &registers, &registers);
 }

void Set_Vga_Mode (void)
 {
  union REGS registers;

  registers.h.ah = 0x00;		/* Servizio 0 dell' interrupt 0x10 */
  registers.h.al = 0x13;		/* Numero del nuovo modo grafico   */
  int386 (0x10, &registers, &registers);
 }

void PutPixel (int x, int y, char color)
 {
  _farnspokeb(x + (y*320), color);
 }

void main (void)
 {
  Create_Selector ();

  Set_Vga_Mode ();	      /* Attiviamo la Vga */
  PutPixel (160, 100, 150);   /* Accendiamo 1 pixel nel mezzo dello schermo */
  getch();		      /* Attendiamo la pressione di un tasto */
  Set_Text_Mode ();           /* Riattiviamo la modalit testuale */
 }
