/*
  Programmazione Grafica 3D - Cap. 2
  by Andrea Fasano

  Nel seguente listato procediamo alla creazione di un campo stellato
  tridimensionale, alla sua proiezione prospettica ed alla relativa
  stampa a monitor.

  convertito per DJGPP da Antonio Bonifati
  website: http://Assembly_Page_di_Ant.freeweb.org
*/

#include <stdio.h>
#include <math.h>
#include "demogfx.h"

#define MAXSTARS    500       /* Numero delle stelle (cambiare a piacere!) */
#define MAXDISTANCE 256       /* Max. distanza sulla z */
#define SCALE       255       /* "Grandezza" del nostro mondo */

typedef struct {
    int x, y, z;              /* Coordinate reali della stella      */
    int xp, yp;               /* Coordinate proiettate della stella */
} TYPE_STARS;

TYPE_STARS starfield[MAXSTARS];

/*-------------------------------------------------------------------------*/
/* FUNZIONE create_single_star                                             */
/* UTILIZZO Crea una singola stella, inizializzando opportunatamente un    */
/*          elemento dell' array starfield                                 */
/* INPUT    n - Indice dell' array starfield                               */
/*          distance - Distanza z ove far "apparire" la stella             */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void create_single_star (int n, int distance)
{

   do {
     starfield[n].x = rand () % 320 - 160;
     starfield[n].y = rand () % 200 - 100;
    } while ((starfield[n].x == 0) && (starfield[n].y == 0));

    starfield[n].z = distance;

    starfield[n].xp = 0;
    starfield[n].yp = 0;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE create_starfield                                               */
/* UTILIZZO Inizializza tutto il campo stellato, riempiendo ogni elemento  */
/*          di starfield                                                   */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void create_starfield (void)
{
  int i;

  for (i=1; i<MAXSTARS; i++)
     create_single_star (i, i);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE project_starfield                                              */
/* UTILIZZO Calcola le coordinate proiettate sul monitor di ogni singola   */
/*          stella, tramite la le legge di proiezione prospettica          */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void project_starfield (void)
{
 int i;

 for (i=1; i<MAXSTARS; i++)
  {
    starfield[i].xp = ((starfield[i].x * SCALE) / starfield[i].z) + 160;
    starfield[i].yp = ((starfield[i].y * SCALE) / starfield[i].z) + 100;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE draw_starfield                                                 */
/* UTILIZZO Stampa sul monitor l' intero campo stellato. Per ogni stella   */
/*          viene effettuato il controllo che essa si trovi all' interno   */
/*          dello schermo visibile, altrimenti vieni fatta "rinascere"     */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void draw_starfield (void)
{
  int i;

  for (i=1; i<MAXSTARS; i++)
   if ((starfield[i].xp>=0) && (starfield[i].xp<320) &&
       (starfield[i].yp>=0) && (starfield[i].yp<200))
      putpixel (starfield[i].xp, starfield[i].yp, 100);
    else
      create_single_star (i, MAXDISTANCE);
}

void main (void)
 {
  Init_Graph_Lib();

  set_vga_mode ();           /* Settiamo la scheda in 320x200 */

  create_starfield ();       /* Creiamo il campo stellato */
  project_starfield ();      /* Ora calcoliamo la sua proiezione */
  draw_starfield ();         /* Stampiamolo sullo schermo */

  getch();                   /* Aspettiamo la pressione di un tasto */
  set_text_mode ();          /* Ooops! Indietro al DOS */

  Close_Graph_Lib();
 }
