/*
  Programmazione Grafica 3D - Cap. 2
  by Andrea Fasano

  Implementiamo la tecnica del double buffering per evitare lo sfarfallio.

  Questo  il nostro primo programma davvero buono!!

  convertito per DJGPP da Antonio Bonifati
  website: http://Assembly_Page_di_Ant.freeweb.org

  Per compilare:
  gcc -s -O3 esempio3.c -o esempio3.exe
*/

#include <stdio.h>
#include <math.h>
#include "demogfx.h"

#define MAXSTARS  10000       /* Numero delle stelle (cambiare a piacere!) */
#define MAXDISTANCE 256       /* Max. distanza sulla z */
#define SCALE       255       /* "Grandezza" del nostro mondo */
#define SCROLL_SPEED  2       /* Velocit di scrolling */

typedef struct {
    int x, y, z;              /* Coordinate reali della stella      */
    int xp, yp;               /* Coordinate proiettate della stella */
} TYPE_STARS;

TYPE_STARS   starfield[MAXSTARS];

/*-------------------------------------------------------------------------*/
/* FUNZIONE prepare_palette                                                */
/* UTILIZZO Crea una palette fatta di sfumature di grigio                  */
/* INPUT    palette - palette da riempire                                  */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void prepare_palette (TYPE_PALETTE *palette)
{
  int j;

  for (j=1; j<256; j++)
      {
       palette[j].r = (256-j) / 4;
       palette[j].g = (256-j) / 4;
       palette[j].b = (256-j) / 4;
      }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE create_single_star                                             */
/* UTILIZZO Crea una singola stella, inizializzando opportunatamente un    */
/*          elemento dell' array starfield                                 */
/* INPUT    n - Indice dell' array starfield                               */
/*          distance - Distanza z ova far "apparire" la stella             */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void create_single_star (int n, int distance)
{

   do {
     starfield[n].x = rand () % 320 - 160;
     starfield[n].y = rand () % 200 - 100;
    } while ((starfield[n].x == 0) && (starfield[n].y == 0));

    starfield[n].z = distance;

    starfield[n].xp = 0;
    starfield[n].yp = 0;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE create_starfield                                               */
/* UTILIZZO Inizializza tutto il campo stellato, riempiendo ogni elemento  */
/*          di starfield                                                   */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void create_starfield (void)
{
  int i;

  for (i=1; i<MAXSTARS; i++)
     create_single_star (i, i);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE project_starfield                                              */
/* UTILIZZO Calcola le coordinate proiettate sul monitor di ogni singola   */
/*          stella, tramite la legge di proiezione prospettica             */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void project_starfield (void)
{
 int i;

 for (i=1; i<MAXSTARS; i++)
  {
    starfield[i].xp = ((starfield[i].x * SCALE) / starfield[i].z) + 160;
    starfield[i].yp = ((starfield[i].y * SCALE) / starfield[i].z) + 100;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE draw_starfield                                                 */
/* UTILIZZO Stampa in una generica area di memoria l'intero campo stellato.*/
/*          Per ogni stella viene effettuato il controllo che essa si trovi*/
/*          all' interno dello schermo visibile, altrimenti vieni fatta    */
/*          "rinascere". Ad ogni stella viene associato un colore          */
/*          dipendente dalla sua attuale posizione z all' interno del campo*/
/*          stellato.                                                      */
/* INPUT    screen - Indirizzo iniziale dello schermo virtuale             */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void draw_starfield (VIRTUAL screen)
{
  int i;
  char star_color;

  for (i=1; i<MAXSTARS; i++)
   if ((starfield[i].xp>=0) && (starfield[i].xp<320) &&
       (starfield[i].yp>=0) && (starfield[i].yp<200))
        {
          if (starfield[i].z > 255) star_color = 255;
          else star_color = starfield[i].z;

          put_pixel (starfield[i].xp, starfield[i].yp, star_color, screen);
        }
    else
      create_single_star (i, MAXDISTANCE);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE move_starfield                                                 */
/* UTILIZZO Decrementa il valore z di ogni stella, facendo s che il campo */
/*          stellato si muova verso l' osservatore. Se tale valore        */
/*          negativo, tale stella viene ricreata.                          */
/* INPUT    step - Velocit di scrolling del campo stellato                */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void move_starfield (int step)
{
  int i;

  for (i=1; i<MAXSTARS; i++)
   {
    if (starfield[i].z > step) starfield[i].z -= step;
    else create_single_star (i, MAXDISTANCE);
   }
}

void main (void)
 {
  Init_Graph_Lib();     /* non dimenticare mai questa istruzione! */

  set_vga_mode ();           /* Settiamo la scheda in 320x200 */

  create_starfield ();       /* Creiamo il campo stellato */

  prepare_palette (pal);     /* Mettiamo su un adeguata palette */
  set_palette (pal);         /* Attiviamola!                    */

  virtual_screen = create_screen (320, 200);  /* Creaimo il nostro schermo  */
                                              /* virtuale, avente dimensioni*/
                                              /* identiche a quello della   */
                                              /* Vga                        */

  while (!kbhit())
   {
        clear_screen (virtual_screen); /* Ripuliamo l' area di lavoro */


        move_starfield (SCROLL_SPEED); /* Muoviamo il campo stellato */
        project_starfield ();          /* Ora calcoliamo la sua proiezione */
        draw_starfield (virtual_screen);  /* Stampiamolo nello schermo */
                                          /* virtuale                  */
        wait_retrace ();
        copy_screen_to_vga (virtual_screen); /* Trasferiamolo sul monitor */
   }

  set_text_mode ();          /* Ooops! Indietro al DOS */

  Close_Graph_Lib();
 }
