/*
 *  Programmazione Grafica 3D - Cap. 2
 *  by Andrea Fasano
 *
 *  Un primo semplice effetto grafico: la neve.
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 *  nota: per me DJGPP  il miglior compilatore del mondo,
 *        ed  FREE !!! Download e informazioni a:
 *
 *                  http://www.delorie.com/djgpp/
 */

#include <stdio.h>
#include "demogfx.h"

#define MAXFLAKES     500  /* Numero dei fiocchi di neve */
#define IS_FALLING      0  /* Il fiocco sta cadendo */
#define IS_BLOCKED      1  /* Il fiocco  bloccato */

typedef struct {
        int startx;     /* Ascissa di partenza */
        int radius;     /* Ampliezza del "volteggiamento" del fiocco */
        int x;          /* Ascissa attuale del fiocco di neve  */
        int y;          /* Ordinata attuale del fiocco di neve */
        int dx;         /* Velocit (in pixel) di spostamento orizzontale */
        int angle;      /* Attuale angolo di spostamento                  */
        char status;    /* Controllo per stabilire se il fiocco di neve deve
                           continuare a cadere oppure no */
} FLAKE;


FLAKE snow[MAXFLAKES];     /* La nostra bufera di neve! */
VIRTUAL mask;              /* Schermo virtuale, contenente l'immagine */
TYPE_PALETTE maskpal[256]; /* La sua palette */


/*-------------------------------------------------------------------------*/
/* FUNZIONE load_raw                                                       */
/* UTILIZZO Carica un file di tipo raw, di uno schermo 320x200x256, prece- */
/*          dentemente salvato. Potete creare tali file con programmi tipo */
/*          l'Autodesk Animator, oppure in giro si trovano numerosi con-   */
/*          vertitori da pcx,iff,gif a raw. NON vi  alcun check di errore */
/* INPUT    filename - Nome del file da caricare                           */
/*          pic      - Schermo virtuale di destinazione. Deve essere       */
/*                     allocato prima!                                     */
/*          pal      - Palette di destinazione                             */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void load_raw (char *filename, VIRTUAL pic, TYPE_PALETTE *pal)
{
 FILE *fp;

 fp = fopen (filename, "rb");
 fread (pic, 64000, 1, fp);
 fread (pal, 768, 1, fp);
 fclose (fp);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE generate_single_flake                                          */
/* UTILIZZO Crea un singolo fiocco di neve.                                */
/* INPUT    Indice dell'array snow                                         */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void generate_single_flake (int index)
{
 snow[index].radius = rand () % 20;
 snow[index].x = snow[index].startx = rand () % 320;
 snow[index].y = rand () % 70;

 snow[index].dx = 0;
 snow[index].status = IS_FALLING;
 snow[index].angle = 0;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE generate_all_snow                                              */
/* UTILIZZO Crea ed inizializza l'intero array snow                        */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void generate_all_snow (void)
{
 int n;

 for (n=0; n<MAXFLAKES; n++)
  generate_single_flake (n);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE move_single_flake                                              */
/* UTILIZZO Sposta un singolo fiocco di neve. Se il fiocco  in caduta li- */
/*          bera, viene fatto cadere con movimento sinusoidale, altrimenti */
/*          il suo spostamento orizzontale  definito dalla funzione move_ */
/*          snow, la quale decide quale "direzione" far prendere al fiocco */
/*          di neve a seconda della maschera presente.                     */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void move_single_flake (int index, char touching)
{
 if (!touching)
  {
   snow[index].angle++;
   /* mai raggiunto */
   /* if (snow[index].angle > 359) snow[index].angle = 0; */
   snow[index].x = snow[index].startx +
                   fixed2int ((sin_table[snow[index].angle] *
                                 snow[index].radius));
  }
 else
  {
   snow[index].angle = 0;
   snow[index].x += snow[index].dx;
   snow[index].startx = snow[index].x;
  }
 snow[index].y++;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE move_snow                                                      */
/* UTILIZZO Sposta tutti i fiocchi di neve, a seconda della maschera. Prima*/
/*          viene controllato che il pixel immediatamente sottostante al   */
/*          fiocco di neve sia "libero" (se il colore nella maschera  0,  */
/*          significa che non vi  niente, altrimenti ogni altro colore   */
/*          interpretato come un ostacolo), altrimenti controlla prima il  */
/*          pixel a destra e poi a sinistra. Se tutte e tre le posizioni   */
/*          sono bloccate, significa che il fiocco non puo' piu' muoversi, */
/*          quindi viene updatata la maschera con la posizione del fiocco, */
/*          e tale fiocco viene fatto "rinascere", in maniera tale da avere*/
/*          una nevicata infinita.                                         */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void move_snow (void)
{
 int n, tmpx, tmpy;
 char next_pixel;

 for (n=0; n<MAXFLAKES; n++)
   {
    /*** Il fiocco sta ancora cadendo? ***/
    if (snow[n].status == IS_FALLING)
     {
       tmpx = snow[n].x;
       tmpy = snow[n].y;

       /*** Controlliamo il pixel sotto il fiocco ***/
       next_pixel = get_pixel (tmpx, tmpy+1, mask);
       if (next_pixel)
         {
           /*** Controlliamo il pixel sotto il fiocco a destra ***/
           next_pixel = (char)  get_pixel (tmpx+1, tmpy+1, mask);
           if (next_pixel)
             {
               /*** Controlliamo il pixel sotto il fiocco a sinistra ***/
               next_pixel = (char) get_pixel (tmpx-1, tmpy+1, mask);
               if (next_pixel)
                 /*** Nulla da fare! ***/
                 snow[n].status = IS_BLOCKED;
               else
                 {
                   /*** A sinistra  libero ***/
                   snow[n].dx = -1;
                   move_single_flake (n, 1);
                 }
             }
           else
             {
               /*** A destra  libero ***/
               snow[n].dx = 1;
               move_single_flake (n, 1);
             }
         }
       else
         {
           /*** Sotto  libero ***/
           snow[n].dx = 0;
           move_single_flake (n, 0);
         }
     }
    else
     {
      /*** Il fiocco non ha piu' vie di scampo. Update della maschera
           e sua rinascita ***/
      put_pixel (snow[n].x, snow[n].y, 255, mask);
      generate_single_flake(n);
     }
   }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE draw_snow                                                      */
/* UTILIZZO Disegna il campo innevato, con clipping.                       */
/* INPUT    Schermo di destinazione                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void draw_snow (VIRTUAL screen)
{
 int i;

 for (i=0; i<MAXFLAKES; i++)
  {
   if ((snow[i].x>=0) && (snow[i].x<320) && (snow[i].y>=0) && (snow[i].y<200))
    put_pixel (snow[i].x, snow[i].y, 255, screen);
   else generate_single_flake (i);
  }
}

void main (void)
{
 Init_Graph_Lib();

 /*** Inizializzazione dello schermo in 320x200x256 ***/
 set_vga_mode ();

 /*** Creazione random della neve ***/
 generate_all_snow ();

 /*** Creazione delle tabelle di lookup ***/
 prepare_lookup_tables();

 /*** Allochiamo memoria per i nostri 2 schermi virtuali ***/
 /*** mask conterr poi l'immagine "Dev vi augura buon natale" ***/
 virtual_screen = create_screen (320, 200);
 mask           = create_screen (320, 200);

 /*** Carichiamo il file raw ***/
 load_raw ("dev.raw", mask, maskpal);
 set_palette (maskpal);

 /*** Stampiamo l'immagine sullo schermo ***/
 copy_screen_to_vga(mask);

 /*** Fintanto che non vi viene a noia... ***/
 while (!kbhit())
  {
   /*** Si muove la neve ***/
   move_snow();

   /*** Si copia la maschera attuale in virtual_screen ***/
   flip_screen (virtual_screen, mask);

   /*** Si disegna la posizione attuale in virtual_screen ***/
   draw_snow (virtual_screen);

   /*** Rallentamento forzato, tarato per un P150 ***/
   delay (50);

   /*** Aspettiamo il retrace ***/
   wait_retrace ();

   /*** Si copia il virtual sul monitor ***/
   copy_screen_to_vga(virtual_screen);
  }

 /*** Ritorno al buon vecchio DOS ***/
 set_text_mode();

 Close_Graph_Lib();
}
