                                /* bitmap.h */

/* aggiunto da Antonio */

/* nota: per semplicit non ho implementato qui il doppio meccanismo di
   accesso alla memoria video; quindi questo funziona solo con DEBUG=0
   (implementatelo voi...)                                              */

/* alcune semplici funzioni che permettono di caricare e salvare bitmap
   uncompresse a 4 o 8 bpp (con un qualsiasi numero di colori fino ovviamente
   a 2^4=16 e 2^8=256 colori rispettivamente), disegnarle sullo schermo, ecc...

   Utile soprattutto la funzione di salvataggio, per esportare fotogrammi
   delle vostre demo nel vostro programma di disegno preferito.

   il formato utilizzato  il Windows' RGB-encoded BMP
   e per semplicit assumiano che la larghezza in pixel di una immagine
   sia un multiplo di 4 (a voi il compito di superare questo limite)
   inoltre per semplicit se la bitmap  a 4 bpp, essa viene comunque
   memorizzata impiegando 8 bpp.

   queste procedure sono basate su quello che ho imparato qui:

                        http://www.brackeen.com/home/vga/

   ovviamente a voi il compito di estendere questo file .h con nuove
   funzionalit.
*/

#ifndef _BITMAP
#define _BITMAP

#include <stdio.h>
/* questo file  una estensione di demogfx.h */
#include "../demogfx.h"

#define min(x,y) ((x<y)?x:y) /* macro per ritornare il
                                minimo di due numeri   */

typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned long  dword;

typedef struct tagBITMAP         /* la struttura che contiene una bitmap. */
{
  word width;
  word height;
  word num_colors;                   /* il numero di colori della palette */
  TYPE_PALETTE palette[256];         /* la palette della bitmap */

  byte *data;
} BITMAP;


/* swap: classica funzione di scambio */
void swap(int *px, int *py)   /* scambia *px cpn *py */
{
  int temp;

  temp = *px;
  *px = *py;
  *py = temp;
}

/**************************************************************************
 *  fskip                                                                 *
 *     Salta un certo numero di byte in un file                           *
 **************************************************************************/

void fskip(FILE *fp, int num_bytes)
{
   int i;
   for (i=0; i<num_bytes; i++)
      fgetc(fp);
}


/**************************************************************************
 *  load_bmp                                                              *
 *    Carica in file bitmap in memoria.                                   *
 **************************************************************************/

void load_bmp(char *file,BITMAP *b)
{
  FILE *fp;
  long index;
  int x, i;
  byte BitsPerPixel;
  byte abyte;

  /* apre il file */
  if ((fp = fopen(file,"rb")) == NULL)
  {
    printf("Error opening file %s.\n",file);
    exit(1);
  }

  /* controlla che si tratti di un file bitmap valido */
  if (fgetc(fp)!='B' || fgetc(fp)!='M')
  {
    fclose(fp);
    printf("%s is not a bitmap file.\n",file);
    exit(1);
  }

  /* leggi solo la larghezza e l'altezza dell'immagine
     il numero di bpp e di colori usati; ignora il resto */
  fskip(fp,16);
  fread(&b->width, sizeof(word), 1, fp);
  fskip(fp,2);
  fread(&b->height,sizeof(word), 1, fp);
  fskip(fp,4);
  fread(&BitsPerPixel, sizeof(byte), 1, fp);

  /* controllo che il BitsPerPixel  uno dei
     due valori supportati */
  if (BitsPerPixel!=4 && BitsPerPixel!=8)
  {
    printf("sorry, this program can't load %d-bpp bitmaps\n", BitsPerPixel);
    exit(1);
  }

  fskip(fp,17);
  fread(&b->num_colors,sizeof(word), 1, fp);
  fskip(fp,6);

  /* se il numero di colori usato non  specificato,
     cio  posto a zero, assumiamolo pari al max
     possibile per questo numero di BitsPerPixel        */
  if (b->num_colors==0)
     switch (BitsPerPixel)
     {
        case 4: b->num_colors=16;  break;     /* 2^4=16  */
        case 8: b->num_colors=256; break;     /* 2^8=256 */
     }

  /* prova ad allocare la memoria necessaria */
  if ((b->data = (byte *) malloc((word)(b->width*b->height))) == NULL)
  {
    fclose(fp);
    printf("Error allocating memory for file %s.\n",file);
    exit(1);
  }

  /* leggi la palette della bitmap */
  /* da notare che solo i colori usati dalla bitmap,
     cio num_colors colori, sono settati;
     gli altri 256-num_colors colori li dovete settare
     voi se vi servono.                                */
  for(index=0;index<b->num_colors;index++)
  {
    b->palette[index].b = fgetc(fp) >> 2;
    b->palette[index].g = fgetc(fp) >> 2;
    b->palette[index].r = fgetc(fp) >> 2;
    x=fgetc(fp);        /* salta byte di riempimento */
  }

  /* leggi la bitmap: da notare che le linee sono memorizzate nel
     file a rovescio: cio la prima linea nel file  l'ultima sullo
     schermo e la lettura si fa in due modi diversi a seconda che
     il numero di bpp  4 o 8 (supportiamo solo questi due casi).
     Inoltre nel caso di bpp 4, per semplicit convertiamo a bpp 8,
     anzich memorizzare due pixel per byte.
     Una cosa da non dimenticare  che ci possono essere dei byte di
     riempimento (padding) che occorre saltare; il formato stabilisce
     infatti che ogni scanline sia un certo numero intero di DWORD
     (1 DWORD=4 bytes).

     Non ho ritenuto necessario ottimizzare questo codice,
     in quanto non  codice che deve essere ripetuto molte volte
     e comunque stiamo leggendo da disco che  molto pi lento
     della CPU. L'ho scritto nel modo pi semplice da scrivere.        */


  for (index=(b->height-1)*b->width; index>=0; index-=b->width)
  {
    switch (BitsPerPixel)
    {
      case 4:
        for (x=0; x<b->width; x+=2)
        {
          abyte=(byte)fgetc(fp);
          b->data[(word)index+x]=abyte>>4;     /* nibble alto  */
          b->data[(word)index+x+1]=abyte&0xF;  /* nibble basso */
        }
        break;

      case 8:
        for (x=0; x<b->width; x++)
          b->data[(word)index+x]=(byte)fgetc(fp);
      break;
    }

    /* salta eventuali bytes di padding (riempimento) */
    for (i=0; i<b->width%4; i++)
      x=fgetc(fp);
  }

  fclose(fp);
}

/**************************************************************************
 *  save_bmp                                                              *
 *    Salva una regione rettangolare dello schermo come bitmap 8bpp       *
 *    (se volete salvare a 4bpp aggiungete questa caratteristica).        *
 *    Questa funzione  molto semplice e salva interamente tutta la       *
 *    palette. Ripeto ancora una volta che questa non funziona per        *
 *    bitmap con 4bpp ma  in grado di scrivere solo bitmap a 8bpp        *
 **************************************************************************/

void save_bmp(char *file, int x1,int y1, int x2, int y2)
{
  FILE *fp;

  word a_word;
  dword a_dword, zero=0;

  TYPE_PALETTE curr_pal[256];
  byte comp, pad;

  dword width, height, size;
  int x,y,i;

  /* apre il file */
  if ((fp = fopen(file,"wb")) == NULL)
  {
    printf("Error writing file %s.\n",file);
    exit(1);
  }

  /* specifica le coordinate del rettangolo come vuoi,
     tanto non mi fai fesso :)                         */
  if (x1>x2) swap(&x1,&x2);
  if (y1>y2) swap(&y1,&y2);

  /* larghezza, altezza, numero di byte di padding alla fine
     di ogni scanline e dimensione della bitmap da salvare   */
  width=(x2-x1+1);
  height=(y2-y1+1);
  pad=width%4;
  /* come potete vedere occorre non dimenticarsi
     di includere nel conto della size eventuali
     byte di padding che inseriremo e il cui numero
      pari a quello delle linee orizzontali che
     compongono la bitmat, cio height per il
     numero di byte di padding alla fine di ogni
     linea (che  zero se width  divisibile per 4):

     width*height+height*pad    matematicamente diventa
     height*(width+pad)                                 */
  size=(width+pad)*height;


  /* i primi due bytes indicano il tipo di file e vanno settati a "BM" */
  fputs("BM",fp);

  /* Dimensione del file (DWORD)                */
  /* la dimensione in byte  la seguente:

                    header di 54-byte    +
        tabella della palette di 1024    +
        la bitmap vera e propria (size)  =
     -------------------------------------
                            1078 + size         */

  a_dword=1078 + size;   /* file size */

  fwrite(&a_dword, sizeof(dword), 1, fp);


  /* Riservato (DWORD). Settare a zero. */
  fwrite(&zero, sizeof(dword), 1, fp);

  /* Offset a cui iniziano i dati della bitmap (DWORD) */
  a_dword=1078;
  fwrite(&a_dword, sizeof(dword), 1, fp);

  /* Dimensione del resto dell'header da questo punto in poi.
     Settare a 40 (DWORD).                                    */
  a_dword=40;
  fwrite(&a_dword, sizeof(dword), 1, fp);

  /* Larghezza della bitmap in pixel (DWORD) */
  fwrite(&width, sizeof(dword), 1, fp);

  /* Altezza della bitmap in pixel (DWORD) */
  fwrite(&height, sizeof(dword), 1, fp);

  /* Numero di Piani. Settare a 1. (WORD) */
  a_word=1;
  fwrite(&a_word, sizeof(word), 1, fp);

  /* Numero di bit per pixel (WORD) */
  a_word=8;
  fwrite(&a_word, sizeof(word), 1, fp);

  /* Compressione. Di solito  sempre 0,
     come dobbiamo impostarlo nel nostro caso (DWORD) */
  fwrite(&zero, sizeof(dword), 1, fp);

  /* Size in bytes of the bitmap (DWORD) */
  fwrite(&size, sizeof(dword), 1, fp);

  /* Pixel Orizzontali per metro (DWORD) */
  /* ci mettiamo semplicemente zero */
  fwrite(&zero, sizeof(dword), 1, fp);

  /* Pixel Verticali per metro (DWORD) */
  /* ci mettiamo semplicemente zero */
  fwrite(&zero, sizeof(dword), 1, fp);

  /* Numero di colori usati (DWORD) */
  /* ci mettiamo sempre 256 (salveremo sempre tutta la palette) */
  a_dword=256;
  fwrite(&a_dword, sizeof(dword), 1, fp);

  /* Numero di colori "importanti" (DWORD) */
  /* per semplicit un 256 va bene */
  a_dword=256;
  fwrite(&a_dword, sizeof(dword), 1, fp);

  /* non dimentichiamoci di salvare la palette! */
  get_palette(curr_pal);  /* legge la palette corrente */
  /* il formato in cui va salvata la palette  strano:
      bgr anzich rgb ed inoltre ogni valore della
     componente di colore va moltiplicato per 4;
     infine (cosa da non dimenticare!) va usato un byte
     di riempimento ogni 3 byte (in modo che ogni palette
     entry sia esattamente una DWORD). Usiamo 0 come byte
     di riempimento.                                      */
  for (i=0; i<256; i++)
  {
     comp=(curr_pal[i].b)<<2;
     fwrite(&comp, sizeof(byte), 1, fp);

     comp=(curr_pal[i].g)<<2;
     fwrite(&comp, sizeof(byte), 1, fp);

     comp=(curr_pal[i].r)<<2;
     fwrite(&comp, sizeof(byte), 1, fp);

     fputc('\0', fp);    /* inserisci byte di riempimento */
  }

  /* eccoci finalmente a salvare la bitmap;
     occorre ricordare che la bitmap va salvata a rovescio! */
  /* attenzione: occorre mettere byte di padding se una linea
     non ha un numero di pixel multipli di 4 in modo da raggiungere
     il successivo multiplo di 4. Anche questo codice non  ottimizzato. */
  for (y=y2; y>=y1; y--)
  {
    for (x=x1; x<=x2; x++)
       fputc(screen[x+y*320],fp);

    for (i=0; i<pad; i++)
       fputc(0, fp);
  }

  fclose(fp); /* amen :) */
}

/**************************************************************************
 *  draw_bitmap                                                           *
 *    Disegna una bitmap.                                                 *
 **************************************************************************/

void draw_bitmap(BITMAP *bmp,int x,int y)
{
  int j;
  word screen_offset = (y<<8)+(y<<6)+x;
  word bitmap_offset = 0;

  for(j=0;j<bmp->height;j++)
  {
    memcpy(&screen[screen_offset],&bmp->data[bitmap_offset],bmp->width);

    bitmap_offset+=bmp->width;
    screen_offset+=320;       /* screen width */
  }
}


/**************************************************************************
 *  draw_transparent_bitmap                                               *
 *    Disegna una bitmap in modo trasparente.                             *
 *    assume come trasparente il colore numero 0 della bitmap             *
 **************************************************************************/

void draw_transparent_bitmap(BITMAP *bmp,int x,int y)
{
  int i,j;
  word screen_offset = (y<<8)+(y<<6);
  word bitmap_offset = 0;
  byte data;

  for(j=0;j<bmp->height;j++)
  {
    for(i=0;i<bmp->width;i++,bitmap_offset++)
    {
      data = bmp->data[bitmap_offset];
      if (data) screen[screen_offset+x+i] = data;
    }
    screen_offset+=320;       /* screen width */
  }
}

/**************************************************************************
 *  draw_bitmap_clip                                                      *
 *    Disegna una bitmap, ma al contrario di draw_bitmap effettua il      *
 *    clipping, cio evita di disegnare fuori dei bordi.                  *
 **************************************************************************/

void draw_bitmap_clip(BITMAP *bmp,int x,int y)
{
  int j, xlim=min(320-x, bmp->width), ylim=min(200-y, bmp->height);
  word screen_offset = (y<<8)+(y<<6)+x;
  word bitmap_offset = 0;

  for(j=0;j<ylim;j++)
  {
    memcpy(&screen[screen_offset],&bmp->data[bitmap_offset],xlim);

    bitmap_offset+=bmp->width;
    screen_offset+=320;       /* screen width */
  }
}

/**************************************************************************
 *  draw_tiled                                                            *
 *    Disegna una bitmap in modo tiled, cio riempie tutto lo schermo     *
 *    affiancando varie copie della bitmap                                *
 *    (effettua il clipping ai bordi destro e inferiore se necessario)    *
 **************************************************************************/

void draw_tiled(BITMAP *bmp)
{
  int x,y;

  for (y=0; y<200; y+=bmp->height)      /* 200=screen height */
    for (x=0; x<320; x+=bmp->width)      /* 320=screen width  */
      draw_bitmap_clip(bmp,x,y);
}

#endif

