/*
 *  Programmazione Grafica 3D - Cap. 3
 *  by Andrea Fasano
 *
 *  In questo secondo codice diamo il movimento al fuoco.
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 */

#include <stdio.h>
#include "demogfx.h"


unsigned char fb[320][200];             /* Il buffer atto a contenere il fuoco */
TYPE_PALETTE firepal[256];     /* La palette del fuoco */


/*-------------------------------------------------------------------------*/
/* FUNZIONE generate_fire_palette                                          */
/* UTILIZZO Setta la palette firepal[] in modo appropriato per "simulare"  */
/*          i colori della fiamma                                          */
/* INPUT    pal - palette da riempire                                      */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void generate_fire_palette (TYPE_PALETTE *pal)
{
 int n;

 /* I primi 64 colori sono organizzati secondo una sfumatura graduale dal
    nero al rosso vivo (il primo colore sar nero) */
 for (n=0; n<64; n++)
  {
   pal[n].r = n;
   pal[n].g = 0;
   pal[n].b = 0;
  }

 /* Altri 64 colori li utilizziamo per le sfumature di giallo della fiamma */
 for (n=64; n<128; n++)
  {
   pal[n].r = 63;
   pal[n].g = n-64;
   pal[n].b = 0;
  }

 /* altri 64 colori li utilizziamo per realizzare
    una sfumatura dal giallo al bianco vivo */
 for (n=128; n<192; n++)
  {
   pal[n].r = 63;
   pal[n].g = 63;
   pal[n].b = n-128;
  }

 /* gli ultimi 64 colori a nostra disponizione vengono utilizzati
    solo per le posizioni pi basse del fuoco, apparendo sotto
    forma di macchioline, e perci ci mettiamo tutti dei bianchi. */
 for (n=192; n<256; n++)
  {
   pal[n].r = 63;
   pal[n].g = 63;
   pal[n].b = 63;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE show_fire                                                      */
/* UTILIZZO Copia il contenuto attuale del buffer fb[][] sullo schermo.    */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void show_fire (void)
{
 int x, y;

   for (x=0; x<320; x++)
    for (y=0; y<200; y++)
       putpixel (x, y, fb[x][y]);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE move_fire                                                      */
/* UTILIZZO Crea il fuoco dal buffer fb e lo "muove" verso l'alto.         */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void move_fire (void)
{
 int x, y;

 for (x=0; x<320; x++)
   for (y=0; y<198; y++)  /** Attenzione all'ultima riga! **/
      {
       /*** Stiamo analizzando il margine sinistro del buffer? ***/
      if (x == 0)
        fb[x][y] = (fb[0][y+1] + fb[x][y+1] + fb[x+1][y+1] + fb[x][y+2])>>2;
      else
        /*** Stiamo analizzando il margine destro del buffer? ***/
        if (x == 319)
          fb[x][y] = (fb[x-1][y+1] + fb[x][y+1] + fb[319][y+1] + fb[x][y+2])>>2;
        else
          /*** Tutto ok ***/
          fb[x][y] = (fb[x-1][y+1] + fb[x][y+1] + fb[x+1][y+1] + fb[x][y+2])>>2;
      }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE update_last_lines                                              */
/* UTILIZZO Riempie le ultime due righe del buffer fb con valori casuali   */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void update_last_lines (void)
{
 int x;

 for (x=0; x<320; x++)
   {
    fb[x][198] = rand () % 256;
    fb[x][199] = rand () % 256;
   }
}

void main (void)
{
 Init_Graph_Lib();

 set_vga_mode ();                   /* Attiviamo la modalit 320x200x256 */

 generate_fire_palette (firepal);   /* Creiamo un'opportuna palette */
 set_palette (firepal);             /* Settiamola */

 memset (fb, 0, 64000);             /* Azzeriamo il buffer */

 while (!kbhit())                   /* Fintanto che un tasto non  premuto */
  {
   update_last_lines ();
   move_fire ();
   show_fire ();
  }

 set_text_mode ();                  /* Oops! Indietro al buon vecchio DOS.. */

 Close_Graph_Lib();
}
