/*
 *  Programmazione Grafica 3D - Cap. 3
 *  by Andrea Fasano
 *
 *  Rendiamo pi veloce il nostro effetto del fuoco.
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 */

#include <stdlib.h>
#include <math.h>
#include "demogfx.h"

#define MAX_HEIGHT 15          /* Max altezza raggiunta dal fuoco */
char FIRE_HEIGHT = 1;          /* Step per la sfumatura verso il nero*/

VIRTUAL double_buffer;
unsigned char fb[320][200+2];  /* Il buffer atto a contenere il fuoco.
                                  Esso  2 linee in pi rispetto alla normale
                                  grandezza dello schermo, in quanto non
                                  verranno mostrate dal momento che sono un
                                  po' bruttine */
TYPE_PALETTE firepal[256];     /* La palette del fuoco */


/*-------------------------------------------------------------------------*/
/* FUNZIONE generate_fire_palette                                          */
/* UTILIZZO Setta la palette firepal[] in modo appropriato per "simulare"  */
/*          i colori della fiamma                                          */
/* INPUT    pal - palette da riempire                                      */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void generate_fire_palette (TYPE_PALETTE *pal)
{
 int n;

 /* I primi 64 colori sono organizzati secondo una sfumatura graduale dal
    nero al rosso vivo (il primo colore sar nero) */
 for (n=0; n<64; n++)
  {
   pal[n].r = n;
   pal[n].g = 0;
   pal[n].b = 0;
  }

 /* Altri 64 colori li utilizziamo per le sfumature di giallo della fiamma */
 for (n=64; n<128; n++)
  {
   pal[n].r = 63;
   pal[n].g = n-64;
   pal[n].b = 0;
  }

 /* altri 64 colori li utilizziamo per realizzare
    una sfumatura dal giallo al bianco vivo */
 for (n=128; n<192; n++)
  {
   pal[n].r = 63;
   pal[n].g = 63;
   pal[n].b = n-128;
  }

 /* gli ultimi 64 colori a nostra disponizione vengono utilizzati
    solo per le posizioni pi basse del fuoco, apparendo sotto
    forma di macchioline, e perci ci mettiamo tutti dei bianchi. */
 for (n=192; n<256; n++)
  {
   pal[n].r = 63;
   pal[n].g = 63;
   pal[n].b = 63;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE move_fire                                                      */
/* UTILIZZO Anima il fuoco di 1 passo, interpolando ogni punto con la      */
/*          media dei quattro punti sottostanti nel buffer fb[][].         */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void move_fire (void)
{
 int x, y;

  for (x=0; x<320; x++)
    for (y=0; y<200; y++)
      {
       /*** Stiamo analizzando il margine sinistro del buffer? ***/
       if (x == 0)
        fb[x][y] = (fb[0][y+1] + fb[x][y+1] + fb[x+1][y+1] + fb[x][y+2])>>2;
       else
        /*** Stiamo analizzando il margine destro del buffer? ***/
        if (x == 319)
         fb[x][y] = (fb[x-1][y+1] + fb[x][y+1] + fb[319][y+1] + fb[x][y+2])>>2;
        else
          /*** Tutto ok ***/
          fb[x][y] = (fb[x-1][y+1] + fb[x][y+1] + fb[x+1][y+1] + fb[x][y+2])>>2;

       /*** Sfumiamo leggermente verso il nero! ***/
       if (fb[x][y] > MAX_HEIGHT) fb[x][y] -= FIRE_HEIGHT;

       double_buffer [x + y*320] = fb[x][y];
      }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE update_last_lines                                              */
/* UTILIZZO Aggiorna le ultime 2 linee del buffer fb[][] con valori casuali*/
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void update_last_lines (void)
{
 int x;

 for (x=0; x<320; x++)
   {
    fb[x][200] = rand () % 256;
    fb[x][201] = rand () % 256;
   }
}

void main (void)
{
 Init_Graph_Lib();

 set_vga_mode ();                   /* Attiviamo la modalit 320x200x256 */

 double_buffer = create_screen (320, 202); /* Allochiamo memoria x il nostro
                                            double buffer */

 generate_fire_palette (firepal);   /* Creiamo un'opportuna palette */
 set_palette (firepal);             /* Settiamola */

 memset (fb, 0, 64000);             /* Azzeriamo il buffer */

 while (!kbhit())                   /* Attendiamo la pressione di un tasto */
  {
   update_last_lines ();
   move_fire ();                    /* Ciclo principale del fuoco */

   copy_screen_to_vga (double_buffer);   /* Copiamo il doppio buffer */
  }

 set_text_mode ();                  /* Oops! Indietro al buon vecchio DOS.. */

 Close_Graph_Lib();
}
