/*
 *  Programmazione Grafica 3D - Cap. 3
 *  by Andrea Fasano
 *
 *  Questo codice si occupa di stampare sul video un'immagine statica
 *  del plasma calcolato.
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 */

#include <stdio.h>
#include <math.h>
#include "demogfx.h"

#define WAV_VELOX_1 1           /* Costanti utilizzate per la velocit di */
#define WAV_VELOX_2 2           /* ogni onda del coseno */
#define WAV_VELOX_3 3
#define WAV_VELOX_4 4

#define WAV_STEP_1 1            /* Costanti utilizzate per l'incremento di */
#define WAV_STEP_2 2            /* velocit di ogni onda */
#define WAV_STEP_3 1
#define WAV_STEP_4 2

TYPE_PALETTE plasmapal[256];    /* La palette del nostro plasma */

/*-------------------------------------------------------------------------*/
/* FUNZIONE generate_plasma_palette                                        */
/* UTILIZZO Setta la palette del plasma...                                 */
/* INPUT    pal - palette da riempire                                      */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void generate_plasma_palette (TYPE_PALETTE *pal)
{
 int n;

 /*** La palette del plasma  costituita essenzialmente da 4 blocchi di
      sfumature continue. Divertitevi pure a cambiarle a piacimento ***/
 for (n=0; n<64; n++)
  {
   pal[n].r = 63-n;
   pal[n].g = n;
   pal[n].b = 0;
  }
 for (n=64; n<128; n++)
  {
   pal[n].r = 63;
   pal[n].g = n-63;
   pal[n].b = 0;
  }
 for (n=128; n<192; n++)
  {
   pal[n].r = 0;
   pal[n].g = n/6;
   pal[n].b = n-128;
  }
 for (n=192; n<256; n++)
  {
   pal[n].r = 63;
   pal[n].g = n-192;
   pal[n].b = 0;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE prepare_plasma_tables                                          */
/* UTILIZZO Memorizza negli array sin_table e cos_table i valori del seno  */
/*          e del coseno moltiplicati e divisi per qualche costante.       */
/*          Notare che tali valori vengono shiftati a sinistra di 16 bit   */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void prepare_plasma_tables (void)
{
 int i;

 for (i=0; i<256; i++)
  {
   /* sin_table non viene usata in questo plasma, ma eccovi la sua
      tabella nel caso vogliate provare a fare il vostro plasma
      personale. */
   sin_table [i] = float2fixed ((float) 30*(cos (i*2 * 3.1415 / 256)));
   cos_table [i] = float2fixed ((float) 60*(sin (i*2 * 3.1415 / 256)));
  }
}


void main ()
{
  /*** Variabili di comodo ***/
  int x, y, plasma_color;
  unsigned char wav1, wav2, wav3, wav4;

  Init_Graph_Lib();

  /*** Precalcoliamoci le nostre tabelline per il seno ed il coseno ***/
  prepare_plasma_tables ();

  /*** VGA 320x200x256 ***/
  set_vga_mode ();
  /*** Mettiamo su una palette un po' strana... ***/
  generate_plasma_palette (plasmapal);
  /*** Attiviamola ***/
  set_palette (plasmapal);

  //view_palette ();    /* per vedere la palette che stiamo
                        /* usando, prima di stampare il plasma */

  /*** Inizializziamo i primi due indici ***/
  wav1 = WAV_VELOX_1;
  wav2 = WAV_VELOX_2;

  for (y=0; y<200; y++)
    {
      /*** Risettiamo correttamente wav3 e wav4 ***/
      wav3 = WAV_VELOX_3;
      wav4 = WAV_VELOX_4;

      for(x=0; x<320; x++)
      {
        /*** Questo  il cuore del calcolo del plasma. Notate l'utilizzo della
             notazione fissa ***/
        plasma_color = fixed2int (
            cos_table[wav1] +           /* provate a usare sempre il seno */
            cos_table[wav2] +           /* a "mischiare" seno e coseno    */
            cos_table[wav3] +
            cos_table[wav4]
           );

        /*** Stampiamo il risultato alle coordinate x, y ***/
        putpixel (x, y, plasma_color);

        /*** Incrementiamo il movimento del plasma ***/
        wav3 += WAV_STEP_3;
        wav4 += WAV_STEP_4;
      }

      /*** Incrementiamo il movimento del plasma ***/
      wav1 += WAV_STEP_1;
      wav2 += WAV_STEP_2;
    }

  /*** Attende la pressione di un qualsivoglia tasto ***/
  getch ();

  /*** Yep! Indietro al DOS... ***/
  set_text_mode ();

  Close_Graph_Lib();
}
