/*
 *  Programmazione Grafica 3D - Cap. 3
 *  by Andrea Fasano
 *
 * - PlasFire -
 * Programma che fonde insieme il plasma ed il fuoco. Qui praticamente
 * stampiamo il plasma "filtrato" dal fuoco. Il codice qui presentato 
 * stato preso dal listato4 e dal listato6
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 *  download GRATIS di DJGPP a:         http://www.delorie.com/djgpp/
 */

#include <stdlib.h>
#include <math.h>
#include "demogfx.h"

/******************* Costanti utilizzate dal plasma ***********************/
#define WAV_VELOX_1 1           /* Costanti utilizzate per la velocit di */
#define WAV_VELOX_2 2           /* ogni onda del coseno */
#define WAV_VELOX_3 3
#define WAV_VELOX_4 4

#define WAV_STEP_1 1            /* Costanti utilizzate per l'incremento di */
#define WAV_STEP_2 2            /* velocit di ogni onda */
#define WAV_STEP_3 1
#define WAV_STEP_4 2

/******************* Costanti utilizzate dal fuoco ************************/
#define MAX_HEIGHT 150         /* Max altezza raggiunta dal fuoco */
char FIRE_HEIGHT = 1;          /* Step per la sfumatura verso il nero*/

VIRTUAL double_buffer;
unsigned char fb[320][203];    /* Il buffer atto a contenere il fuoco. Esso 
                                 2 linee in pi rispetto alla normale
                                 grandezza dello schermo, in quanto non verranno
                                 mostrate dal momento che sono un po' bruttine */
TYPE_PALETTE firepal[256];     /* La palette del fuoco, ivi utilizzata diversamente */


/*-------------------------------------------------------------------------*/
/* FUNZIONE move_fire                                                      */
/* UTILIZZO Anima il fuoco di 1 passo, interpolando ogni punto con la media*/
/*          dei quattro punti sottostanti nel buffer fb[][]                */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void move_fire (void)
{
 int x, y, n=0;
 char *ptr = (char *) double_buffer;

  for (x=0; x<320; x++)
    for (y=0; y<200; y++)
      {
       /*** Stiamo analizzando il margine sinistro del buffer? ***/
       if (x == 0)
        fb[x][y] = (fb[0][y+1] + fb[x][y+1] + fb[x+1][y+1] + fb[x][y+2])>>2;
       else
        /*** Stiamo analizzando il margine destro del buffer? ***/
        if (x == 319)
         fb[x][y] = (fb[x-1][y+1] + fb[x][y+1] + fb[319][y+1] + fb[x][y+2])>>2;
        else
          /*** Tutto ok ***/
          fb[x][y] = (fb[x-1][y+1] + fb[x][y+1] + fb[x+1][y+1] + fb[x][y+2])>>2;

       /*** Sfumiamo leggermente verso il nero! ***/
       if (fb[x][y] > MAX_HEIGHT) fb[x][y] -= FIRE_HEIGHT;

       /** Filtraggio **/
       double_buffer [x + y*320] -= fb[x][y];
      }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE generate_strange_palette                                       */
/* UTILIZZO Setta la palette con colori scelti abbastanza casualmente.     */
/* INPUT    pal - palette da riempire                                      */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void generate_strange_palette (TYPE_PALETTE *pal)
{
 int n;

 for (n=0; n<64; n++)
  {
   pal[n].r = 0;
   pal[n].g = 0;
   pal[n].b = n;
  }
 for (n=64; n<128; n++)
  {
   pal[n].r = 0;
   pal[n].g = n-64;
   pal[n].b = n-64;
  }
 for (n=128; n<192; n++)
  {
   pal[n].r = n-128;
   pal[n].g = 0;
   pal[n].b = n-128;
  }
 for (n=192; n<256; n++)
  {
   pal[n].r = n-192;
   pal[n].g = n-192;
   pal[n].b = n-192;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE update_last_lines                                              */
/* UTILIZZO Aggiorna le ultime 2 linee del buffer fb[][] con valori casuali*/
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void update_last_lines (void)
{
 int x;

 for (x=0; x<320; x++)
   {
    fb[x][200] = rand () % 256;
    fb[x][201] = rand () % 256;
   }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE prepare_plasma_tables                                          */
/* UTILIZZO Memorizza negli array sin_table e cos_table i valori del seno  */
/*          e del coseno moltiplicati e divisi per qualche costante.       */
/*          Notare che tali valori vengono shiftati a sinistra di 16 bit   */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void prepare_plasma_tables (void)
{
 int i;

 for (i=0; i<256; i++)
  {
   sin_table [i] = float2fixed ((float) 10*(cos (i*2 * 3.1415 / 256)));
   cos_table [i] = float2fixed ((float) 30*(sin (i*2 * 3.1415 / 256)));
  }
}


/*-------------------------------------------------------------------------*/
/* FUNZIONE do_plasma                                                      */
/* UTILIZZO Crea il plasma e lo stampa nella variabile di destinazione dest*/
/* INPUT    dest - schermo virtuale di destinazione                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
unsigned char wav_vlx1 = 0, wav_vlx2 = 0, wav_vlx3 = 0, wav_vlx4 = 0;
unsigned char wav1, wav2, wav3, wav4;
void do_plasma (char *dest)
{
 /*** Variabili di comodo utilizzate dalla funzione do_plasma ***/
 int x, y, plasma_color;

    /*** Assegniamo i nuovi valori a wav1 e wav2 ***/
    wav1 = wav_vlx1;
    wav2 = wav_vlx2;

    for (y=0; y<200; y++)
    {
      /*** Assegniamo i nuovi valori a wav3 e wav4 ***/
      wav3 = wav_vlx3;
      wav4 = wav_vlx4;

      for(x=0; x<320; x++)
      {
        /*** Questo  il cuore del calcolo del plasma. Notate l'utilizzo della
             notazione fissa ***/
        plasma_color = fixed2int (
            cos_table[wav1] +
            cos_table[wav2] +
            cos_table[wav3] +
            cos_table[wav4]
           );

        /*** Stampiamo il risultato alle coordinate x, y ***/
        dest[x+y*320]=plasma_color;

        /*** Incrementiamo il movimento del plasma ***/
        wav3 += WAV_STEP_3;
        wav4 += WAV_STEP_4;
      }

      /*** Incrementiamo il movimento del plasma ***/
      wav1 += WAV_STEP_1;
      wav2 += WAV_STEP_2;
    }

    /*** Una volta terminato di disegnare una schermata, aggiorniamo il
         movimento ***/
    wav_vlx1 += WAV_VELOX_1;
    wav_vlx2 -= WAV_VELOX_2;
    wav_vlx3 += WAV_VELOX_3;
    wav_vlx4 -= WAV_VELOX_4;
  }


void main (void)
{
 Init_Graph_Lib();

 set_vga_mode ();                   /* Attiviamo la modalit 320x200x256 */

 double_buffer = create_screen (320, 205); /* Allochiamo memoria x il nostro
                                            double buffer */

 prepare_plasma_tables ();          /* Generiamo le tabelle di look up */
 generate_strange_palette (firepal); /* Creiamo una palette un po' diversa */
 set_palette (firepal);             /* Settiamola */

 //view_palette();      /* vediamo che tipo di palette stiamo usando */

 while (!kbhit())                   /* Attendiamo la pressione di un tasto */
  {
   update_last_lines ();
   do_plasma (double_buffer);       /* Riempiamo double_buffer con il plasma*/
   move_fire ();                    /* Ciclo principale del fuoco */
   copy_screen_to_vga (double_buffer);
  }

 set_text_mode ();                  /* Oops! Indietro al buon vecchio DOS.. */

 Close_Graph_Lib();
}
