/*
 *  Programmazione Grafica 3D - Cap. 3
 *  by Andrea Fasano
 *
 *  analogo a esempio6.c ma con la stessa palette di esempio7.c
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 */

#include <stdio.h>
#include <math.h>
#include "demogfx.h"

#define WAV_VELOX_1 1           /* Costanti utilizzate per la velocit di */
#define WAV_VELOX_2 2           /* ogni onda del coseno */
#define WAV_VELOX_3 3
#define WAV_VELOX_4 4

#define WAV_STEP_1 1            /* Costanti utilizzate per l'incremento di */
#define WAV_STEP_2 2            /* velocit di ogni onda */
#define WAV_STEP_3 1
#define WAV_STEP_4 2

TYPE_PALETTE plasmapal[256];    /* La palette del nostro plasma */

/*-------------------------------------------------------------------------*/
/* FUNZIONE generate_plasma_palette                                        */
/* UTILIZZO Setta la palette del plasma...                                 */
/* INPUT    pal - palette da riempire                                      */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void generate_plasma_palette (TYPE_PALETTE *pal)
{
 int n;

 /*** La palette del plasma  costituita essenzialmente da 4 blocchi di
      sfumature continue. Divertitevi pure a cambiarle a piacimento ***/
 for (n=0; n<64; n++)
  {
   pal[n].r = 0;
   pal[n].g = 0;
   pal[n].b = n;
  }
 for (n=64; n<128; n++)
  {
   pal[n].r = 0;
   pal[n].g = n-64;
   pal[n].b = n-64;
  }
 for (n=128; n<192; n++)
  {
   pal[n].r = n-128;
   pal[n].g = 0;
   pal[n].b = n-128;
  }
 for (n=192; n<256; n++)
  {
   pal[n].r = n-192;
   pal[n].g = n-192;
   pal[n].b = n-192;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE prepare_plasma_tables                                          */
/* UTILIZZO Memorizza negli array sin_table e cos_table i valori del seno  */
/*          e del coseno moltiplicati e divisi per qualche costante.       */
/*          Notare che tali valori vengono shiftati a sinistra di 8 bit    */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void prepare_plasma_tables (void)
{
 int i;

 for (i=0; i<256; i++)
  {
   sin_table [i] = float2fixed ((float) 30*(cos (i*2 * 3.1415 / 256)));
   cos_table [i] = float2fixed ((float) 60*(sin (i*2 * 3.1415 / 256)));
  }
}


void main ()
{
  /*** Variabili di comodo ***/
  int x, y, plasma_color;
  /************************************ vanno unsigned ???? */
  unsigned char wav_vlx1 = 0, wav_vlx2 = 0, wav_vlx3 = 0, wav_vlx4 = 0;
  unsigned char wav1, wav2, wav3, wav4;

  Init_Graph_Lib();

  /*** Precalcoliamoci le nostre tabelline per il seno ed il coseno ***/
  prepare_plasma_tables ();

  /*** VGA 320x200x256 ***/
  set_vga_mode ();
  /*** Mettiamo su una palette un po' strana... ***/
  generate_plasma_palette (plasmapal);
  /*** Attiviamola ***/
  set_palette (plasmapal);

  //view_palette();

  /*** Fintanto che un tasto non viene premuto... ***/
  while (!kbhit ())
  {
    /*** Assegniamo i nuovi valori a wav1 e wav2 ***/
    wav1 = wav_vlx1;
    wav2 = wav_vlx2;

    for (y=0; y<200; y++)
    {
      /*** Assegniamo i nuovi valori a wav3 e wav4 ***/
      wav3 = wav_vlx3;
      wav4 = wav_vlx4;

      for(x=0; x<320; x++)
      {
        /*** Questo  il cuore del calcolo del plasma. Notate l'utilizzo della
             notazione fissa ***/
        plasma_color = fixed2int (
            cos_table[wav1] +    /* potete cambiare un po' il tipo di  */
            cos_table[wav2] +    /* di plasma semplicemente alternando */
            cos_table[wav3] +    /* seni e coseni!!                    */
            cos_table[wav4]
           );

        /*** Stampiamo il risultato alle coordinate x, y ***/
        putpixel (x, y, plasma_color);

        /*** Incrementiamo il movimento del plasma ***/
        wav3 += WAV_STEP_3;
        wav4 += WAV_STEP_4;
      }

      /*** Incrementiamo il movimento del plasma ***/
      wav1 += WAV_STEP_1;
      wav2 += WAV_STEP_2;
    }

    /*** Una volta terminato di disegnare una schermata, aggiorniamo il
         movimento ***/
    wav_vlx1 += WAV_VELOX_1;
    wav_vlx2 -= WAV_VELOX_2;
    wav_vlx3 += WAV_VELOX_3;
    wav_vlx4 -= WAV_VELOX_4;
  }

  /*** Yep! Indietro al DOS... ***/
  set_text_mode ();

  Close_Graph_Lib();
}
