/*
 *  Programmazione Grafica 3D
 *  by Andrea Fasano
 *
 *  Libreria per la creazione di demo nella modalit standard 320x200x256.
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 *  per segnalazioni di bug, o richieste di nuove funzioni da implementare
 *  scrivetemi all'indirizzo che trovate nel mio sito web.
 */

#ifndef _DEMOGFX_
#define _DEMOGFX_

#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include <dos.h>
#include "vaccess.h"

/****************************************************************************/
/*  DEFINIZIONE DEI TIPI                                                    */
/****************************************************************************/
typedef unsigned char * VIRTUAL;
typedef struct {              /* Questa struttura ci serve per definire */
   char r, g, b;              /* la nostra palette.                     */
} TYPE_PALETTE;

/****************************************************************************/
/*  VARIABILI GLOBALI                                                       */
/****************************************************************************/
VIRTUAL  virtual_screen;  /* Il nostro schermo virtuale, dove           */
                          /* indirizzeremo tutte le operazioni grafiche */
TYPE_PALETTE pal[256];    /* la palette dello schermo */

int sin_table[360], cos_table[360]; /* Le nostre tabelle di lookup */

/****************************************************************************/
/*  PROTOTIPI                                                               */
/****************************************************************************/
void Init_Graph_Lib(void);      /* chiamare prima di tutto (obbligatoria) */
void Close_Graph_Lib(void);

void set_vga_mode  (void);
void set_text_mode (void);

void putpixel (int x, int y, char color);
char getpixel (int x, int y);
void clearscreen (void);
void putvertline (int y1, int y2, int x, char color);

void put_pixel          (int x, int y, char color, VIRTUAL screen);
char get_pixel (int x, int y, VIRTUAL screen_buf);
void clear_screen       (VIRTUAL screen);
void copy_screen_to_vga (VIRTUAL screen);
void flip_screen (VIRTUAL dest, VIRTUAL source);

void set_color   (char num_color, char red, char green, char blue);
TYPE_PALETTE get_color (char num_color);
void set_palette (TYPE_PALETTE *palette);
void get_palette (TYPE_PALETTE *palette);
void view_palette (void);

void wait_retrace          (void);
VIRTUAL create_screen      (int x_size, int y_size);
void prepare_lookup_tables (void);
void load_raw (char *filename, VIRTUAL screen, TYPE_PALETTE *palette, int size);

int   int2fixed   (int x);
int   fixed2int   (int x);
float fixed2float (int x);
int   float2fixed (float x);

/****************************************************************************/
/*  FUNZIONI                                                                */
/****************************************************************************/
/*-------------------------------------------------------------------------*/
/* FUNZIONE set_vga_mode                                                   */
/* UTILIZZO Inizializza la scheda grafica nella modalit 320x200           */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_vga_mode(void)
{
  union REGS registers;

  registers.h.ah = 0x00;
  registers.h.al = 0x13;
  int386 (0x10, &registers, &registers);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_text_mode                                                  */
/* UTILIZZO Inizializza la scheda grafica nella modalit testuale          */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_text_mode(void)
{
  union REGS registers;

  registers.h.ah = 0x00;
  registers.h.al = 0x03;
  int386 (0x10, &registers, &registers);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE putpixel                                                       */
/* UTILIZZO Stampa un pixel sullo schermo                                  */
/* INPUT    x - ascissa del punto                                          */
/*          y - ordinata del punto                                         */
/*          color - colore del punto                                       */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void putpixel (int x, int y, char color)
{
  #if DEBUG
     _farnspokeb(x + (y<<8)+(y<<6), color);
  #else
     vga_screen[x+ (y<<8)+(y<<6)]=color;
  #endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE put_pixel                                                      */
/* UTILIZZO Stampa un pixel in una generica area di memoria                */
/*          con le ottimizzazioni attivate  velocissima con DJGPP, non    */
/*          c' bisogno di codice assembly                                 */
/* INPUT    x - ascissa del punto                                          */
/*          y - ordinata del punto                                         */
/*          color - colore del punto                                       */
/*          screen_buf - Indirizzo iniziale dello schermo virtuale         */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void put_pixel (int x, int y, char color, VIRTUAL screen_buf)
{
   screen_buf[x+(y<<8)+(y<<6)]=color;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE putvertline                                                    */
/* UTILIZZO Plotta una linea verticale                                     */
/* INPUT    x - ascissa della linea                                        */
/*          y1 - ordinata iniziale                                         */
/*          y2 - ordinata finale                                           */
/*          color - colore della linea                                     */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void putvertline (int y1, int y2, int x, char color)
 {
  int h, n, start;

  /* Calcoliamo l' altezza della linea, ovvero il
      numero di pixel da stampare */
  h = abs (y2 - y1) + 1;

  if (y1>y2)
      y1=y2;

  /* Calcoliamo l' indirizzo iniziale del primo punto */
  start = (x + (y1*320));    /* DJGPP mette da s gli shifting
                                se le ottimizzazioni sono attivate */
  for (n=0;n<h;n++)
  {
     #if DEBUG
       _farnspokeb(start, color);
     #else
       vga_screen[start]=color;
     #endif
     start+=320;
  }
 }

/*-------------------------------------------------------------------------*/
/* FUNZIONE clear_screen                                                   */
/* UTILIZZO Cancella lo schermo di una generica area di memoria            */
/* INPUT    screen - Indirizzo iniziale dello schermo virtuale da ripulire */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void clear_screen(VIRTUAL screen_buf)
{
  memset(screen_buf,0,64000);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE clearscreen                                                    */
/* UTILIZZO Cancella lo schermo del monitor                                */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void clearscreen (void)
{
  #if DEBUG
     int i;
     for (i=0; i<16000; i++)
        _farnspokel(i<<2, 0);
  #else
     int i;
     for (i=0; i<16000; i++)
        ((int*)vga_screen)[i]=0; /* azzera 4 byte per volta con accesso diretto */
  #endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE copy_screen_to_vga                                             */
/* UTILIZZO Trasferisce il contenuto della generica area di memoria        */
/*          sul monitor                                                    */
/* INPUT    screen_buf - Indirizzo iniziale dello schermo virtuale         */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void copy_screen_to_vga (VIRTUAL screen_buf)
 {
  #if DEBUG
     movedata(_my_ds(), (int)screen_buf, video, 0, 320*200);
  #else
     int i;
     for (i=0; i<16000; i++)
        ((int*)vga_screen)[i]=((int*)screen_buf)[i];
  #endif
 }

/*-------------------------------------------------------------------------*/
/* FUNZIONE flip_screen                                                    */
/* UTILIZZO Copia il contenuto dello schermo virtuale source nello schermo */
/*          virtuale dest                                                  */
/* INPUT    source - Schermo virtuale sorgente                             */
/*          dest   - Schermo virtuale destinatario                         */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void flip_screen (VIRTUAL dest, VIRTUAL source)
{
   memcpy(dest,source,64000);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE get_pixel                                                      */
/* UTILIZZO Ritorna il colore di uno schermo virtuale                      */
/* INPUT    x - ascissa del punto                                          */
/*          y - ordinata del punto                                         */
/*          screen_buf - schermo virtuale sorgente                         */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
char get_pixel (int x, int y, VIRTUAL screen_buf)
{
  return screen_buf[x+ (y<<8)+(y<<6)];
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE getpixel                                                       */
/* UTILIZZO Ritorna il colore della vga                                    */
/* INPUT    x - ascissa del punto                                          */
/*          y - ordinata del punto                                         */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
char getpixel (int x, int y)
{
  #if DEBUG
     return _farnspeekb(x + (y<<8)+(y<<6));
  #else
     return vga_screen[x+ (y<<8)+(y<<6)];
  #endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_color                                                      */
/* UTILIZZO Setta i valori delle componenti rosso, verde e blue di un      */
/*          singolo colore memorizzate nell' array pal della VGA           */
/* INPUT    num_color - Indice del colore nell' array pal                  */
/*          red - Valore della componente rossa del colore                 */
/*          green - Valore della componente verde del colore               */
/*          blue - Valore della componente blue del colore                 */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_color (char num_color, char red, char green, char blue)
 {
  outp (0x3c8, num_color);
  outp (0x3c9, red);
  outp (0x3c9, green);
  outp (0x3c9, blue);
 }

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_palette                                                    */
/* UTILIZZO Attiva sullo schermo i colori memorizzati nella palette        */
/* INPUT    palette - palette da attivare                                  */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_palette (TYPE_PALETTE *palette)
 {
  int i;

  for (i=0; i<256; i++)
    set_color (i, palette[i].r, palette[i].g, palette[i].b);
 }

/*-------------------------------------------------------------------------*/
/* FUNZIONE get_color                                                      */
/* UTILIZZO E' l'opposto di set_color. Restituisce i valori delle          */
/*          componenti rosso, verde e blue di un singolo colore            */
/*          memorizzate nell' array palette della VGA                      */
/* INPUT    num_color - Indice del colore nell' array pal                  */
/* OUTPUT   struct TYPE_PALETTE                                            */
/*-------------------------------------------------------------------------*/
TYPE_PALETTE get_color (char num_color)
 {
  TYPE_PALETTE palette;

  outp (0x3c7, num_color);
  palette.r = inp (0x3c9);
  palette.g = inp (0x3c9);
  palette.b = inp (0x3c9);

  return palette;
 }

/*-------------------------------------------------------------------------*/
/* FUNZIONE get_palette                                                    */
/* UTILIZZO Opposta di set_palette. Restituisce nell'array palette la      */
/*          palette della VGA correntemente impostata.                     */
/* INPUT    Nessunopa                                                      */
/* OUTPUT   nel parametro palette                                          */
/*-------------------------------------------------------------------------*/
void get_palette (TYPE_PALETTE *palette)
 {
  int i;

  for (i=0; i<256; i++)
    palette[i] = get_color (i);
 }

/*-------------------------------------------------------------------------*/
/* FUNZIONE view_palette                                                   */
/* UTILIZZO Visualizza la palette corrente "in forma grafica", sotto forma */
/*          di 256 linee verticali. Dapprima cancella lo schermo, poi      */
/*          attende un tasto e lo cancella di nuovo (se non vi va bene     */
/*          questo cambiatela!)                                            */
/*          E' utile per verificare che le vostre funzioni per settare la  */
/*          palette agiscano come effettivamente volete. Quindi questa     */
/*          funzione serve nelle demo esclusivamente a scopo di verifica.  */
/*          Utilizzo tipico:                                               */
/*                                                                         */
/*                chiamata a procedura che crea una palette                */
/*                pal in memoria e la setta                                */
/*                view_palette(pal);                                       */
/*                codice di inizio della demo                              */
/*                                                                         */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void view_palette (void)
{
   int i;

   /* cancella lo schermo - se non volete cancellarlo toglietela */
   clearscreen();

   /* visualizza la palette sotto forma di banda continua colorata */
   for (i=0; i<256; i++)
      putvertline (0, 199, 32+i, i);  /* aggiungo 32 per centrare la banda */

   /* attendi un tasto */
   getch();

   /* cancella di nuovo lo schermo */
   clearscreen();
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE wait_retrace                                                   */
/* UTILIZZO Attende il ritracciamento verticale del pennello elettronico   */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void wait_retrace (void)
 {
  while ( (inp (0x3da) & 0x8) != 0 ) {};
  while ( (inp (0x3da) & 0x8) == 0 ) {};
 }

/*-------------------------------------------------------------------------*/
/* FUNZIONE create_screen                                                  */
/* UTILIZZO Crea un' area di memoria destinata ad essere uno "schermo      */
/*          virtuale"                                                      */
/* INPUT    x_size - Larghezza dell' area                                  */
/*          y_size - Altezza dell' area                                    */
/* OUTPUT   Puntatore all' area creata                                     */
/*-------------------------------------------------------------------------*/
VIRTUAL create_screen (int x_size, int y_size)
{
 return (VIRTUAL) malloc (x_size * y_size);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE prepare_lookup_tables                                          */
/* UTILIZZO Memorizza negli array sin_table e cos_table i valori del seno  */
/*          e del coseno che saranno utilizzati per la rotazione del       */
/*          campo stellato, o in altre demo.                               */
/*          Notare che tali valori vengono shiftati a sinistra di 8 bit    */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void prepare_lookup_tables (void)
{
 int i;

 for (i=0; i<360; i++)
  {
   sin_table [i] = float2fixed ((float) (sin (i * 3.1415 / 180)));
   cos_table [i] = float2fixed ((float) (cos (i * 3.1415 / 180)));
  }
}


/* Queste sono funzioni di conversione da notazione floating point a
   notazione fixed point del tipo 24.8
   Per avere la notazione 16.16 basta cambiare 8 con 16 e 256.0
   con 65536.0 e analogamente per avere altri tipi di notazione fissa.
   In generale per la notazione 32-n.n basta shiftare di n posti
   fixed2int e int2fixed e moltiplicare/dividere per 2^n in
   float2fixed e fixed2float. Non scordatevi il .0 nella costante
   in fixed2float! - oppure cambiate il casting cos:
                     return ((float) x / 256);

   Se dovete cambiare spesso il tipo di notazione da un programma
   all'altro definitevi (con define) delle costanti. */

/*-------------------------------------------------------------------------*/
/* FUNZIONE fixed2int                                                      */
/* UTILIZZO Converte una variabile fixed in intera                         */
/* INPUT    x - variabile in notazione fissa                               */
/* OUTPUT   valore intero                                                  */
/*-------------------------------------------------------------------------*/
int   fixed2int (int x)
{
 return (x >> 8);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE int2fixed                                                      */
/* UTILIZZO Converte una variabile intera in fixed                         */
/* INPUT    x - variabile intera                                           */
/* OUTPUT   valore in notazione fissa                                      */
/*-------------------------------------------------------------------------*/
int   int2fixed (int x)
{
 return (x << 8);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE float2fixed                                                    */
/* UTILIZZO Converte una variabile float in notazione fissa                */
/* INPUT    x - variabile float                                            */
/* OUTPUT   valore in notazione fissa                                      */
/*-------------------------------------------------------------------------*/
int   float2fixed (float x)
{
 return ((int) (x * 256.0));
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE fixed2float                                                    */
/* UTILIZZO Converte una variabile fixed in una variabile float            */
/* INPUT    x - variabile in notazione fissa                               */
/* OUTPUT   valore float                                                   */
/*-------------------------------------------------------------------------*/
float fixed2float (int x)
{
 return ((float) (x / 256.0));
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE load_raw                                                       */
/* UTILIZZO Carica dal disco un file .raw (non vi  alcun error checking)  */
/* INPUT    filename - Nome del file da caricare                           */
/*          screen   - Puntatore dell'area di memoria di destinazione.     */
/*                     Attenzione! Abbiate cura di allocare sufficiente    */
/*                     spazio per screen, prima di utilizzare tae funzione */
/*          palette  - Palette di destinazione                             */
/*          size     - Grandezza del file da leggere - 768                 */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void load_raw (char *filename, VIRTUAL screen, TYPE_PALETTE *palette, int size)
{
 FILE *fp;

 fp = fopen (filename, "rb");    /* Apertura del file */
 fseek (fp, -768L, SEEK_END);    /* Cerchiamo in fondo al file la palette */
 fread (palette, 768, 1, fp);    /* Leggiamola */
 fseek (fp, 0, SEEK_SET);        /* Si riparte dall'inizio del file */
 fread (screen, size, 1, fp);    /* Lettura del file */
 fclose (fp);
}

#endif
