/*
 *  Programmazione Grafica 3D - Cap. 4
 *  by Andrea Fasano
 *
 * Libreria per la gestione delle modalit x 320x200x256 con 4 pagine di
 * memoria grafica e 320x240x256 con 3,4 pagine di memoria grafica.
 *
 * Nota: ho convertito questi programmi sotto DJGPP per me stesso
 *       e li metto a disposizione di tutti. Ringrazio l'autore
 *       per la sua disponibilit.
 *       Anche perch sono un principiante, non offro nessuna
 *       garanzia di funzionamento, n posso essere ritenuto
 *       responsabile di danni, ecc... insomma il solito
 *       "use at you own risk". Commenti, suggerimenti, critiche,
 *       ecc.. sono ben accetti. Questa libreria  allo stato
 *       sperimentale - se vi servono altre funzioni grafiche
 *       non incluse fatemi sapere e probabilmente le implementer.
 *       Nonostante non ho avuto il tempo di effettuare controlli
 *       accuratissimi di correttezza, i programmi del libro di
 *       Andrea inclusi hanno girato tutti correttamente, perci
 *       questa libreria dovrebbe essere abbastanza corretta
 *       (tranquilli, non penso proprio che brucerete il monitor!)
 *       Se avete problemi fatemi sapere.
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 *
 * Nota sulla protezione della memoria:
 *      (e per chi vuole implementare meccanismi pi raffinati)
 *       come implementare la protezione della memoria nel modo x?
 *       ecco alcune note e la scelta che ho fatto:
 *       la prima cosa da osservare  che, facendo come nella
 *       libreria demogfx.h, a causa dell'organizzazione della
 *       memoria della vga in piani di bit, la scrittura su
 *       una coordinata non troppo alta rispetto ai massimi
 *       consentiti pu anche non generare alcun errore e andare
 *       a finire su un'altra pagina!!
 *       Ad esempio, plottando nel modo x 320x200 un pixel a
 *       (320,199) sulla pagina 0, in realt verr plottato
 *       (0,0) sulla pagina 1 senza alcun errore !! (vedi put_pixel_x)
 *       In effetti in questo caso non si  fatta alcuna
 *       violazione di accesso alla memoria: abbiamo sempre
 *       acceduto entro i limiti del segmento in cui la vga
 *       ci fa vedere il bit-plane. Questo fatto pu (dico pu) essere
 *       comunque considerato un errore di tipo logico,
 *       quindi per scoprirlo potreste semplicemente aggiungere
 *       controlli di clipping alle varie routine qui presenti
 *       dove necessario (ad es. in put_pixel_x). La cosa
 *       migliore sarebbe aggiungere un altro flag dipendente
 *       da DEBUG che, se DEBUG  attivato, seleziona e deseleziona
 *       tali controlli di clipping. Cos l'utente pu scegliere
 *       se la scrittura nella pagina/e successiva va vista come
 *       errore oppure no impostando o deselezionando questo flag
 *       come facevamo col flag DEBUG. Io per semplicit non
 *       ho implementato questo, ma non dovrebbe essere difficile
 *       aggiungerlo se vi sembra utile.
 *       Un'altra cosa  che un selettore largo 64000 come in
 *       demogfx.h non permette di accedere alla memoria fuori
 *       schermo nella modalit 320x200 e non  una buona misura
 *       per la modalit 320x240; non volendo accedere nemmeno
 *       ad 1 byte fuori schermo in tal caso ci vorrebbe un
 *       selettore di 57600; in pratica la cosa migliore sarebbe
 *       introdurre un ulteriore flag che permette di scegliere
 *       se abilitare l'accesso alla memoria fuori schermo (che
 *       allunga l'ultima pagina rispetto alle altre). Non abilitare
 *       l'accesso alla memoria fuori schermo in 320x200 significa
 *       avere un selettore per 64000 byte, mentre in 320x240
 *       occorre un selettore di soli 57600.
 *       Con un selettore cos, anche se non  attivato il flag
 *       che d errore se da una pagina cerco di scrivere in
 *       un'altra, la scrittura fuori schermo, come ad esempio
 *       plottare (320,199) sull'ultima pagina in 320x200 (la numero 3)
 *       otterrete un errore GPF e la sua diagnostica,  causa il
 *       superamento di 64000 bytes.
 *       (ovviamente stiamo parlando del caso in cui DEBUG=1)
 *      
 *       In questa versione ho semplicemente definito un
 *       selettore di 64kb, anzich definire due selettori;
 *       comunque in vaccess2 trovate un esempio di come
 *       definire due selettori distinti.
 *       Usare un solo selettore  la soluzione pi semplice
 *       se si vuole sempre permettere l'accesso alla memoria
 *       fuori schermo (cos si hanno 3,4 pagine circa in
 *       320x240 anzich il numero tondo 3 e poche righe in
 *       pi in 320x200). Ricordo che 320x240  la risoluzione
 *       x a pixel quadrati. Per maggiori dettagli vedi file
 *       vaccess.h.
 *
 *       il file demogfx.h incluso in questo capitolo utilizza
 *       vaccess.h, quindi lo stesso selettore da 64Kb per la
 *       modalit 320x200 standard. Ci significa che potete
 *       accedere alla memoria fuori schermo (dopo l'offset 63999
 *       fino a 65535 compreso). Se non volete permettere questo
 *       accesso in 320x200, allora avete bisogno di selettori
 *       separati e funzioni di inizializzazione separate; per
 *       semplicit (e visto che ho poco tempo per scrivere e
 *       settare questo codice), ho usato un unico selettore,
 *       che comunque assicura una buona protezione della
 *       memoria da eventuali "puntatori selvaggi".
 *
 *       Quindi riassumendo: se ad es. plotto un pixel in (320,199)
 *       in un primo caso essendo la pagina attiva la numero 0
 *       e in un secondo caso essendo la pagina attiva la numero 1,
 *       come si dovrebbe comportare la libreria?
 *       Nel primo caso non c' alcun errore (se non un errore di
 *       tipo logico, se volete), poich la scrittura non 
 *       fuori schermo, mentre nel secondo caso s. Se 
 *       importante evitare errori di scrittura in una delle
 *       pagine successive, potete aggiungere alle funzioni di
 *       questa libreria (ad es. a put_pixel_x) i controlli clipping.
 *       Consiglio di farlo nel solo caso in cui
 *       DEBUG  attivato, oppure create un altro flag
 *       per la compilazione condizionale di questi controlli
 *       come sottoflag di DEBUG (ossia agisce solo quando DEBUG=1)
 *       in modo da poterli includere/escludere durante il debugging.
 *       Potreste anche mettere il controllo di clipping
 *       (meglio se sotto compilazione condizionale) nel
 *       vostro codice, ma  noioso doverlo ripetere in
 *       ogni programma.
 *       Ehi, se estendete la libreria, anche in qualsiasi
 *       altro modo, mandatemi la vostra versione via e-mail.
 *
 */

#ifndef _DEMOX_
#define _DEMOX_

#include <stdlib.h>
#include <dos.h>
#include "vaccess.h"    /* per l'accesso alla memoria video
                           in modo protetto (DJGPP compila
                           sempre in modo protetto)
                        */

/****************************************************************************/
/*  VARIABILI GLOBALI                                                       */
/****************************************************************************/
int screen_width,          /* Lunghezza dello schermo */
    screen_height,         /* Altezza dello schermo */
    bitplane_width,        /* Lunghezza di un bit-plane */
    active_page;           /* Numero della pagina attiva */

/****************************************************************************/
/*  PROTOTIPI                                                               */
/****************************************************************************/
void set_modex_320x200 (void);
void set_modex_320x240 (void);

void set_visible_page (int page);
void set_active_page  (int page);
void view_at          (int y_coord);

void put_pixel_x (int x, int y, char color);
char get_pixel_x (int x, int y);

/****************************************************************************/
/*  FUNZIONI                                                                */
/****************************************************************************/
/*-------------------------------------------------------------------------*/
/* FUNZIONE set_modex_320x200                                              */
/* UTILIZZO Inizializza la scheda grafica nella modalit x 320x200         */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_modex_320x200 (void)
{
 union REGS registers;
 #if DEBUG
    int i;
 #endif

 /* Settiamo la scheda nella modalit grafica 320x200x256 *standard* */
 registers.w.ax = 0x13;
 int386 (0x10, &registers, &registers);

 /* Adesso disattiviamo il Chain-4 bit del Sequncer Address Register */
 /* Siamo cos in una modalit unchained */
 outp (0x3c4, 0x04);           /* Porta 0x3c4, Indice 0x04 */
 outp (0x3c5, 0x06);           /* Settiamo i bit appropriati */

 /* Ora disattiviamo la modalit doubleword, nell'UnderLine Location  */
 /* Register */
 outp (0x3d4, 0x14);           /* Porta 0x3d4, Indice 0x14 */
 outp (0x3d5, 0x00);

 /* Disabilitiamo la modalit word, nel Mode Control Register */
 outp (0x3d4, 0x17);           /* Porta 0x3d4, Indice 0x17 */
 outp (0x3d5, 0xe3);

 /* Adesso ripuliamo l'intera memoria grafica, cos da non avere "garbage" */
 /* sullo schermo. Attiviamo contemporaneamente tutti i bit-plane, per far */
 /* prima! */
 outp (0x3c4, 0x02);
 outp (0x3c5, 0x0F);

 #if DEBUG
   for (i=0; i<64000; i+=4)
      _farnspokel(i, 0);

   /* oppure: */
   /* for (i=0; i<16000; i++)
         _farnspokel(i<<2, 0);*/
 #else
   memset (vga_screen, 0, 65536);
 #endif

 /* Ora settiamo adeguatamente le nostre variabili globali */
 screen_width  = 320;
 screen_height = 200;
 bitplane_width = screen_width / 4;

 /* Infine settiamo come default di pagina attiva e visibile la pagina 0 */
 set_visible_page (0);
 set_active_page (0);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_modex_320x240                                              */
/* UTILIZZO Inizializza la scheda grafica nella modalit x 320x240         */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_modex_320x240 (void)
{
 /* Prima attiviamo la modalit x 320x200, in seguito ne modifichiamo la */
 /* altezza, operando con i soliti registri hardware della VGA */
 set_modex_320x200 ();

 outp  (0x3C2, 0xE3);

 /* Si disabilita il write protect bit */
 outpw (0x3D4, 0x2C11);

 /* Settaggio del Vertical Total */
 outpw (0x3D4, 0x0D06);

 /* Overflow register */
 outpw (0x3D4, 0x3E07);

 /* Settaggio del Vertical Retrace Start */
 outpw (0x3D4, 0xEA10);

 /* Settaggio del Start Vertical Blanking */
 outpw (0x3D4, 0xE715);

 /* Settaggio del End Vertical Blanking */
 outpw (0x3D4, 0x0616);

 /* Settaggio del Vertical Retrace End */
 outpw (0x3D4, 0xAC11);

 /* Settaggio del Vertical Display Enable */
 outpw (0x3D4, 0xDF12);

 /* Ora settiamo adeguatamente le nostre variabili globali */
 screen_width  = 320;
 screen_height = 240;
 bitplane_width = screen_width / 4;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_active_page                                                */
/* UTILIZZO Seleziona la pagina attiva dove redirigere tutte le operazioni */
/*          grafiche                                                       */
/* INPUT    page - numero della pagina da attivare (0, 1, 2, 3)            */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_active_page (int page)
{
 active_page = page * bitplane_width * screen_height;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_visible_page                                               */
/* UTILIZZO Seleziona la pagina visibile da mostrare sullo schermo         */
/* INPUT    page - numero della pagina da mostrare (0, 1, 2, 3)            */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_visible_page (int page)
{
 int index;

 index = page * bitplane_width * screen_height;

 outp (0x3d4, 0x0c);
 outp (0x3d5, index >> 8);    /* Mandiamo il byte pi alto */

 outp (0x3d4, 0x0d);
 outp (0x3d5, index & 0xff);  /* Mandiamo il byte pi basso */
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE put_pixel_x                                                    */
/* UTILIZZO Stampa un pixel nella pagina attualmente attiva                */
/* INPUT    x - ascissa del punto                                          */
/*          y - ordinata del punto                                         */
/*          color - colore del punto                                       */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void put_pixel_x (int x, int y, char color)
{
 int index;

 outp (0x3c4, 0x02);
 outp (0x3c5, 0x01 << (x % 4));

 index = (x >> 2) + (y * bitplane_width) + active_page;
 #if DEBUG
    _farnspokeb(index, color);
 #else
    vga_screen [index] = color;
 #endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE get_pixel_x                                                    */
/* UTILIZZO Legge un pixel dalla pagina attualmente attiva                 */
/* INPUT    x - ascissa del punto                                          */
/*          y - ordinata del punto                                         */
/* OUTPUT   colore del punto                                               */
/*-------------------------------------------------------------------------*/
char get_pixel_x (int x, int y)
{
 int index;

 outp (0x3ce, 0x04);  /* Read Plane Select Register */
 outp (0x3cf, x % 4); /* Piano da attivare in lettura */

 index = (x >> 2) + (y * bitplane_width) + active_page;
 #if DEBUG
    return _farnspeekb(index);
 #else
    return vga_screen [index];
 #endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE view_at                                                        */
/* UTILIZZO Mostra il contenuto della memoria grafica da un determinato    */
/*          punto. Non  necessario selezionare la pagina visibile         */
/* INPUT    y_coord - riga di partenza                                     */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void view_at (int y_coord)
{
 int index;

 index = y_coord * bitplane_width;

 outp (0x3d4, 0x0c);
 outp (0x3d5, index >> 8);    /* Mandiamo il byte pi alto */

 outp (0x3d4, 0x0d);
 outp (0x3d5, index & 0xff);  /* Mandiamo il byte pi basso */
}

#endif
