/*
 *  Programmazione Grafica 3D - Cap. 4
 *  by Andrea Fasano
 *
 * Libreria per la gestione delle modalit x 320x200x256 con 4 pagine di
 * memoria grafica e 320x240x256 con 3,4 pagine di memoria grafica.
 *
 * versione per vaccess2.h
 * non testata, inclusa per futuri enhancement della libreria
 * per avere una pi accurata protezione della memoria.
 * un'altra modifica va fatta qui in demogfx.h di questo capitolo:

void copy_screen_to_vga (VIRTUAL screen_buf)
 {
  #if DEBUG
     movedata(_my_ds(), (int)screen_buf, video_200, 0, 320*200); <--- qui
                                                                  video_200
                                                                al posto di
                                                                  video
 *
 */

#ifndef _DEMOX_
#define _DEMOX_

#include <stdlib.h>
#include <dos.h>
#include "vaccess2.h"   /* per l'accesso alla memoria video
                           in modo protetto (DJGPP compila
                           sempre in modo protetto)
                        */

/****************************************************************************/
/*  VARIABILI GLOBALI                                                       */
/****************************************************************************/
int screen_width,          /* Lunghezza dello schermo */
    screen_height,         /* Altezza dello schermo */
    bitplane_width,        /* Lunghezza di un bit-plane */
    active_page;           /* Numero della pagina attiva */

/****************************************************************************/
/*  PROTOTIPI                                                               */
/****************************************************************************/
void set_modex_320x200 (void);
void set_modex_320x240 (void);

void set_visible_page (int page);
void set_active_page  (int page);
void view_at          (int y_coord);

void put_pixel_x (int x, int y, char color);
char get_pixel_x (int x, int y);

/****************************************************************************/
/*  FUNZIONI                                                                */
/****************************************************************************/
/*-------------------------------------------------------------------------*/
/* FUNZIONE set_modex_320x200                                              */
/* UTILIZZO Inizializza la scheda grafica nella modalit x 320x200         */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_modex_320x200 (void)
{
 union REGS registers;
 #if DEBUG
    int i;

    _farsetsel(video_200);
 #endif

 /* Settiamo la scheda nella modalit grafica 320x200x256 *standard* */
 registers.w.ax = 0x13;
 int386 (0x10, &registers, &registers);

 /* Adesso disattiviamo il Chain-4 bit del Sequncer Address Register */
 /* Siamo cos in una modalit unchained */
 outp (0x3c4, 0x04);           /* Porta 0x3c4, Indice 0x04 */
 outp (0x3c5, 0x06);           /* Settiamo i bit appropriati */

 /* Ora disattiviamo la modalit doubleword, nell'UnderLine Location  */
 /* Register */
 outp (0x3d4, 0x14);           /* Porta 0x3d4, Indice 0x14 */
 outp (0x3d5, 0x00);

 /* Disabilitiamo la modalit word, nel Mode Control Register */
 outp (0x3d4, 0x17);           /* Porta 0x3d4, Indice 0x17 */
 outp (0x3d5, 0xe3);

 /* Adesso ripuliamo l'intera memoria grafica, cos da non avere "garbage" */
 /* sullo schermo. Attiviamo contemporaneamente tutti i bit-plane, per far */
 /* prima! */
 outp (0x3c4, 0x02);
 outp (0x3c5, 0x0F);

 #if DEBUG
   for (i=0; i<64000; i+=4)
      _farnspokel(i, 0);

   /* oppure: */
   /* for (i=0; i<16000; i++)
         _farnspokel(i<<2, 0);*/
 #else
   memset (vga_screen, 0, 65536);
 #endif

 /* Ora settiamo adeguatamente le nostre variabili globali */
 screen_width  = 320;
 screen_height = 200;
 bitplane_width = screen_width / 4;

 /* Infine settiamo come default di pagina attiva e visibile la pagina 0 */
 set_visible_page (0);
 set_active_page (0);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_modex_320x240                                              */
/* UTILIZZO Inizializza la scheda grafica nella modalit x 320x240         */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_modex_320x240 (void)
{
 #if DEBUG
    _farsetsel(video_240);
 #endif

 /* Prima attiviamo la modalit x 320x200, in seguito ne modifichiamo la */
 /* altezza, operando con i soliti registri hardware della VGA */
 set_modex_320x200 ();

 outp  (0x3C2, 0xE3);

 /* Si disabilita il write protect bit */
 outpw (0x3D4, 0x2C11);

 /* Settaggio del Vertical Total */
 outpw (0x3D4, 0x0D06);

 /* Overflow register */
 outpw (0x3D4, 0x3E07);

 /* Settaggio del Vertical Retrace Start */
 outpw (0x3D4, 0xEA10);

 /* Settaggio del Start Vertical Blanking */
 outpw (0x3D4, 0xE715);

 /* Settaggio del End Vertical Blanking */
 outpw (0x3D4, 0x0616);

 /* Settaggio del Vertical Retrace End */
 outpw (0x3D4, 0xAC11);

 /* Settaggio del Vertical Display Enable */
 outpw (0x3D4, 0xDF12);

 /* Ora settiamo adeguatamente le nostre variabili globali */
 screen_width  = 320;
 screen_height = 240;
 bitplane_width = screen_width / 4;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_active_page                                                */
/* UTILIZZO Seleziona la pagina attiva dove redirigere tutte le operazioni */
/*          grafiche                                                       */
/* INPUT    page - numero della pagina da attivare (0, 1, 2, 3)            */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_active_page (int page)
{
 active_page = page * bitplane_width * screen_height;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_visible_page                                               */
/* UTILIZZO Seleziona la pagina visibile da mostrare sullo schermo         */
/* INPUT    page - numero della pagina da mostrare (0, 1, 2, 3)            */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_visible_page (int page)
{
 int index;

 index = page * bitplane_width * screen_height;

 outp (0x3d4, 0x0c);
 outp (0x3d5, index >> 8);    /* Mandiamo il byte pi alto */

 outp (0x3d4, 0x0d);
 outp (0x3d5, index & 0xff);  /* Mandiamo il byte pi basso */
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE put_pixel_x                                                    */
/* UTILIZZO Stampa un pixel nella pagina attualmente attiva                */
/* INPUT    x - ascissa del punto                                          */
/*          y - ordinata del punto                                         */
/*          color - colore del punto                                       */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void put_pixel_x (int x, int y, char color)
{
 int index;

 outp (0x3c4, 0x02);
 outp (0x3c5, 0x01 << (x % 4));

 index = (x >> 2) + (y * bitplane_width) + active_page;
 #if DEBUG
    _farnspokeb(index, color);
 #else
    vga_screen [index] = color;
 #endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE get_pixel_x                                                    */
/* UTILIZZO Legge un pixel dalla pagina attualmente attiva                 */
/* INPUT    x - ascissa del punto                                          */
/*          y - ordinata del punto                                         */
/* OUTPUT   colore del punto                                               */
/*-------------------------------------------------------------------------*/
char get_pixel_x (int x, int y)
{
 int index;

 outp (0x3ce, 0x04);  /* Read Plane Select Register */
 outp (0x3cf, x % 4); /* Piano da attivare in lettura */

 index = (x >> 2) + (y * bitplane_width) + active_page;
 #if DEBUG
    return _farnspeekb(index);
 #else
    return vga_screen [index];
 #endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE view_at                                                        */
/* UTILIZZO Mostra il contenuto della memoria grafica da un determinato    */
/*          punto. Non  necessario selezionare la pagina visibile         */
/* INPUT    y_coord - riga di partenza                                     */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void view_at (int y_coord)
{
 int index;

 index = y_coord * bitplane_width;

 outp (0x3d4, 0x0c);
 outp (0x3d5, index >> 8);    /* Mandiamo il byte pi alto */

 outp (0x3d4, 0x0d);
 outp (0x3d5, index & 0xff);  /* Mandiamo il byte pi basso */
}

#endif
