/*
 *  Programmazione Grafica 3D - Cap. 4
 *  by Andrea Fasano
 *
 * In questo esempio vediamo la differenza di risoluzione tra due modalit x,
 * la 320x200 e la 320x240, tramite il disegno di una medesima figura
 * (un cerchio). Notate quanta marcata sia la differenza.
 *
 */

#include <stdio.h>
#include "demogfx.h"
#include "demo_x.h"
#include "circle.h"

/*-------------------------------------------------------------------------*/
/* FUNZIONE draw_circle                                                    */
/* UTILIZZO Disegna un cerchio nella schermo                               */
/*          questa tecnica non  molto precisa e veloce                    */
/*          (non sfrutta la simmetria ottupla), ma tanto per scrivere      */
/*          un codice semplice...                                          */
/* INPUT    x_start - ascissa del centro                                   */
/*          y_start - ordinata del centro                                  */
/*          radius  - raggio                                               */
/*          color   - colore del cerchio                                   */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void draw_circle (int x_start, int y_start, int radius, char color)
{
 int n, x, y;

 for (n=0; n<360; n++)
   {
    x = fixed2int (cos_table[n] * radius);
    y = fixed2int (sin_table[n] * radius);
    put_pixel_x (x_start + x, y_start + y, color);
   }
}

void main (void)
{
 Init_Graph_Lib();

 /* Si precalcola il seno ed il coseno */
 prepare_lookup_tables ();

 /* Attiviamo la modalit x 320x200 con 4 pagine di memoria */
 set_modex_320x200 ();

 /* Disegniamo un bel cerchio nel mezzo dello schermo */
 //draw_circle (160, 100, 50, 100);
 /* l'algoritmo in circle.h  quello di Bresenham; */
 /*  veloce e preciso e non ha bisogno di tabelle di look-up. */
 Circle(160, 100, 50, 100);

 /* Attendiamo la simpatica pressione di un qualsiasi tasto */
 getch();

 /* Attiviamo la modalit x 320x240 con 3,3 pagine di memoria */
 set_modex_320x240 ();

 /* Disegniamo un bel cerchio nel mezzo dello schermo */
 //draw_circle (160, 120, 50, 100);
 /* l'algoritmo in circle.h  quello di Bresenham;  veloce e preciso */
 Circle(160, 120, 50, 100);

 /* Come sopra */
 getch();

 /* Ancora al buon vecchio DOS... */
 set_text_mode ();

 Close_Graph_Lib();
}
