/*
 *  Programmazione Grafica 3D - Cap. 5
 *  by Andrea Fasano
 *
 * Load & Display, per mostrare a video un'immagine in formato PCX.
 *
 * Versione 2 (di Antonio Bonifati)
 *
 * ho migliorato l'interfaccia e risolto alcuni problemi
 * (errori di mancata allocazione in pcx.h e potenziamento di
 *  copy_screen_to_vga per essere in grado di copiare anche
 *  immagini di dimensione minore di tutto lo schermo;
 *  inoltre c'era un problema con le immagini per cui
 *  BytesPerLine non era pari a Xmax-Xmin+1, che non erano
 *  visualizzate correttamente. In generale BytesPerLine
 *   pari a:
 *                    Xmax-Xmin+1   se questo numero  pari
 *      BytesPerLine=
 *                    Xmax-Xmin+2   altrimenti
 *
 *  Quindi BytesPerLine  sempre un numero pari. Lo scopo
 *   quello di avere linee che hanno tutte una lunghezza
 *  pari, in modo che si possano copiare nella memoria
 *  video mediante un numero "tondo" di trasferimenti di 2
 *  byte alla volta. Quando Xmax-Xmin+1  dispari, viene
 *  aggiunto un byte di riempimento (di solito 0) alla
 *  fine di ogni riga di lunghezza Xmax-Xmin+1, in modo
 *  che abbia una lunghezza pari, uguale proprio a
 *  BytesPerLine. Con una semplice modifica in pcx.h
 *  il programma gestisce correttamente anche questo
 *  caso).
 *
 *  Nota: DJGPP fornisce lo stesso supporto per le
 *  wildcard delle shell Unix, quindi ci sono alcune
 *  differenze con il DOS ( anche pi potente).
 *  Ecco degli esempi:
 *               display *.pcx
 *  visualizzare tutti i file .pcx della directory
 *  corrente (basta premere un tasto per visualizzare
 *  il file successivo). L'estensione deve essere
 *  indicata, come pure nell'esempio seguente:
 *              display imgs??.pcx io mygirl
 *  Altri esempi:
 *              display 0*.pcx
 *              display .../*.pcx
 *
 *  Il primo visualizza tutti i file che iniziano per 0 e con
 *  estensione .pcx (anche qui  meglio indicare l'estensione)
 *  il secondo tutti i file pcx in tutte le sottodirectory
 *  a partire da quella corrente, ricorsivamente.
 *
 *  per maggiori dettagli vedi la sezione 16.1
 *  delle faq di djgpp
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 */

#include <stdio.h>
#include <string.h>
#include <keys.h>
#include "demogfx.h"
#include "pcx.h"

/* ritorna 1 se il nome del file non ha estensione, 0 altrimenti */
/* supporta anche estensioni pi lunghe di 3 caratteri e file    */
/* aventi solo l'estensione! (ad es. un file di nome ".prova")   */
int no_extension(char *filename)
{
   int pos=strlen(filename);

   while (pos>0)
     switch (filename[--pos])
       {
         case '.' : return 0;
         case '\\': case '/': return 1;
       }
   return 1;
}

void main (int argc, char *argv[])
{
 VIRTUAL picture;
 TYPE_PALETTE picture_palette[256];
 char file_name[128];
 int i, info;

 if (argc==1)
 {
  set_text_mode ();
  printf("DISPLAY - a simple 320x200x256 colors PCX display\n\n"
         "Syntax: display [-i] filename[.pcx] ...\n"
         "        you can use wildcards like in Unix\n"
         "        -i prints pcx info (of the last file viewed) on exit\n");
  exit(0);
 }

 if (strcasecmp(argv[1], "-i")==0 && argc>2)
 {
    info=1;
    i=2;
 }
 else
 {
    info=0;
    i=1;
 }

 Init_Graph_Lib();

 for (; i<argc; i++) /* per ogni file che soddisfa la wildcard */
 {
    strcpy(file_name, argv[i]);

    if (no_extension(file_name))
       strcat(file_name, ".pcx");

    if ((picture = load_pcx (file_name, picture_palette)) == NULL)
    {
       set_text_mode ();
       printf ("\nImpossibile trovare il file %s o file non valido o pcx non supportato",
               file_name);
       exit (1);
    }

    set_vga_mode ();
    set_palette (picture_palette);

    copy_screen_to_vga (picture, PCX_XSIZE, PCX_YSIZE);

    if (getch()==K_Escape)
       break;
 }

 set_text_mode ();

 if (info)
    print_pcx_info(file_name);

 Close_Graph_Lib();
}
