/*
 *  Programmazione Grafica 3D - Cap. 5
 *  by Andrea Fasano
 *
 *  Converte un file PCX in un file .h
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "demogfx.h"
#include "pcx.h"

void main (int argc, char *argv[])
{
 char *file_pcx, *file_h,
      *nome_pic = "pic000",                     /* Nomi di default */
      *nome_pal = "pal000",                     /*  "   "   "      */
      *pic_initial_string,
      *pal_initial_string,
      *ending_string = "\n};\n",
      *cifra,
      *red,
      *green,
      *blue,
      *cr = "\n",
      *parenthesis_on  = "{",
      *parenthesis_off = "}",
      *comma           = ", ",
      line_count = 0;

 VIRTUAL image;
 TYPE_PALETTE imagepal[256];

 FILE *out;
 int  index;

 file_pcx = argv[1];
 file_h   = argv[2];

 /* L'utente ha scritto qualcosa ? */
 if ((file_pcx == NULL) || (file_h == NULL))
  {
   printf ("\nUso: pcx2arr [file.pcx] [file.h]\n");
   exit (1);
  }

 /* Carichiamo il file pcx e apriamo il file .h */
 if ((image = load_pcx (file_pcx, imagepal)) == NULL)
  {
   printf ("\nImpossibile trovare il file %s o non  un pcx o  un pcx non supportato", file_pcx);
   exit (1);
  }
 out = fopen (file_h, "wb");

 /*** Inizio processo immagine ***/
 printf ("\nInizio processo immagine del file %s...", file_pcx);

 /*** Scriviamo nome e tipo ***/
 pic_initial_string = (char *) malloc (52);
 strcpy (pic_initial_string, "char ");
 strcat (pic_initial_string, nome_pic);
 strcat (pic_initial_string, " [] = {\n");
 fwrite (pic_initial_string, strlen (pic_initial_string), 1, out);

 cifra = (char *) malloc (3);

 for (index=0; index<PCX_SIZE; index++)
  {
   /* Leggiamo i valori dell'immagine, e convertiamoli in ASCII */
   itoa (image[index], cifra, 10);

   /* Scriviamo nel file */
   fwrite (cifra, strlen (cifra), 1, out);

   /* Se siamo in fondo non si deve scrivere la virgola */
   if (index < PCX_SIZE-1) fwrite (comma, strlen(comma), 1, out);

   /* Check per evitare conflitti con vari text-editor */
   line_count += strlen (cifra);
   if (line_count > 75)
    {
     fwrite (cr, strlen (cr), 1, out);
     line_count = 0;
    }
  }
 fwrite (ending_string, strlen (ending_string), 1, out);
 printf (" Fatto.");

 //-----------------------------------------------------------------------//

 /*** Inizio processo palette ***/
 printf ("\nInizio processo palette del file %s...", file_pcx);

 /*** Scriviamo nome e tipo ***/
 pal_initial_string = (char *) malloc (52);
 strcpy (pal_initial_string, "\nTYPE_PALETTE ");
 strcat (pal_initial_string, nome_pal);
 strcat (pal_initial_string, " [] = {\n");
 fwrite (pal_initial_string, strlen (pal_initial_string), 1, out);

 red   = (char *) malloc (3);
 green = (char *) malloc (3);
 blue  = (char *) malloc (3);

 line_count = 0;
 for (index=0; index<256; index++)
  {
   /* Leggiamo le componenti di un colore dalla palette e convertiamole*/
   itoa (imagepal[index].r, red, 10);
   itoa (imagepal[index].g, green, 10);
   itoa (imagepal[index].b, blue, 10);

   /* Adesso scriviamo la tripletta con tutti gli elementi sintattici
      necessari */
   fwrite (parenthesis_on,  strlen(parenthesis_on), 1,  out);
   fwrite (red,   strlen(red),   1, out);
   fwrite (comma, strlen(comma), 1, out);
   fwrite (green, strlen(green), 1, out);
   fwrite (comma, strlen(comma), 1, out);
   fwrite (blue,  strlen(blue),  1, out);
   fwrite (parenthesis_off, strlen(parenthesis_off), 1, out);

   /* In fondo al file la virgola non deve essere messa */
   if (index < 255) fwrite (comma, strlen(comma), 1, out);

   /* Check per evitare conflitti con vari text-editor */
   line_count += strlen (red) + strlen (green) + strlen (blue);
   if (line_count > 30)
    {
     fwrite (cr, strlen (cr), 1, out);
     line_count = 0;
    }
  }

 /* Parentesi finale */
 fwrite (ending_string, strlen (ending_string), 1, out);
 printf (" Fatto.\n");

 fclose (out);
}
