/*
 *  Programmazione Grafica 3D - Cap. 6
 *  by Andrea Fasano
 *
 * In questo esempio vediamo come "catturare" il font di sistema e stampare
 * un messaggio nella modalit grafica 0x13 (320x200x256).
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 */

#include <stdio.h>
#include <conio.h>
#include "demogfx.h"
         #include "go32.h"   /* DJGPP */
         #include <sys/farptr.h>

#define X_CHAR_SIZE 8            /* La dimensione orizzontale del carattere */
#define Y_CHAR_SIZE 8            /* La dimensione verticale del carattere */
#define CHAR_LENGHT 8            /* "Lunghezza" in byte di ogni carattere */
#define NUM_CHAR    256          /* Il numero max. dei caratteri */

/* Questo  il nostro nuovo tipo: una matrice per contenere la forma di ogni
   carattere */
typedef unsigned char TYPE_CHAR [Y_CHAR_SIZE][X_CHAR_SIZE];

unsigned char *vga_text  = (unsigned char *) 0xffa6e;  /* Indirizzo del font*/
char          CHAR_COLOR = 100;  /* Il colore da assegnare ad ogni carattere*/
TYPE_CHAR     alfa[NUM_CHAR];                           /* Il nostro "font" */

/*-------------------------------------------------------------------------*/
/* FUNZIONE get_system_font                                                */
/* UTILIZZO Riempie un array di tipo TYPE_CHAR con i caratteri di sistema  */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void get_system_font (void)
{
 int x, y;            /* Variabili di comodo */
 unsigned char *tmp,  /* Puntatore temporaneo */
          char_row;   /* Una "striscia" (1 byte) del carattere da leggere */
 short index=0;       /* Variabile d'incremento */
 unsigned short old_selector;      /* DJGPP */

 /*
   NB: In memoria, ogni carattere di sistema  memorizzato come una matrice
       di 8x8 bit, per un totale di 64 bit. Ci equivale ad avere una serie
       di 8 byte posti uno dopo l'altro, ed ognuno definisce una "striscia"
       del carattere in questione. Per ogni byte letto dunque, dobbiamo
       controllarne ogni singolo bit: se esso  settato, noi memorizziamo
       un 1 nell'equivalente posizione della matrice alfa[index] prima
       definita, altrimenti vi poniamo uno 0.
 */

 /* Leggiamo il max. numero di caratteri */
 tmp = vga_text + index * CHAR_LENGHT;

 /* settiamo il selettore _dos_ds una volta per tutte invece
    di passarlo ogni volta alla funzione _farpeekb (DJGPP)
    ma prima salviamo quello corrente per poterlo poi
    ripristinare!                                               */
 old_selector=_fargetsel();
 _farsetsel(_dos_ds);

 while (index < NUM_CHAR)       /* il carattero numero 255  compreso */
 {                              /* di solito  un carattere "nero" */
  /* Adesso dobbiamo leggere 64 bit */
  for (y=0; y<Y_CHAR_SIZE; y++)
    {
     /* Memorizziamo in un char (byte) la singola riga del carattere appena
        letto */
     //char_row = tmp[y];
     /* in DJGPP (che usa sempre il modo protetto) si pu fare ad es. cos: */
     /*   (il casting  opzionale) */
     char_row=_farnspeekb((unsigned long)&tmp[y]);

     for (x=0; x<X_CHAR_SIZE; x++)
      {
       /* L'ultimo bit  settato? */
       if (char_row & 0x80)
         alfa[index][x][y] = 1;
       else
         alfa[index][x][y] = 0;

       /* Shiftiamo il byte per leggere il successivo bit*/
       char_row = char_row << 1;
      }
    }
  /* Incrementiamo il contatore di comodo */
  index++;

  /* Calcoliamo l'indirizzo del carattere desiderato */
  tmp+=CHAR_LENGHT;
 }
 _farsetsel(old_selector);      /* ripristina selettore precedente */
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE print_system_char                                              */
/* UTILIZZO Stampa un carattere valido sullo schermo                       */
/* INPUT    startx     - Ascissa iniziale                                  */
/*          starty     - Ordinata iniziale                                 */
/*          ascii_char - Carattere da stampare                             */
/*          color      - Colore del carattere                              */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void print_system_char (int startx, int starty, unsigned char ascii_char)
{
 unsigned short x, y,                   /* Variabili di comodo */
                x_coord, y_coord;       /* Coordinate effettive */
 char           pixel;                  /* Singolo pixel di un carattere */

 for (y=0; y<Y_CHAR_SIZE; y++)
   for (x=0; x<X_CHAR_SIZE; x++)
     {
      x_coord = x + startx;    /* Ascissa reale */
      y_coord = y + starty;    /* Ordinata reale */

      /* Eseguiamo il clipping */
      if ((x_coord < 320) && (y_coord < 200))
       {
        pixel = alfa[ascii_char][x][y];
        if (pixel > 0)
           putpixel (x_coord, y_coord, CHAR_COLOR);
       }
     }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE print_system_string                                            */
/* UTILIZZO Stampa una stringa sullo schermo                               */
/* INPUT    startx - Ascissa iniziale                                      */
/*          starty - Ordinata iniziale                                     */
/*          *msg   - Messaggio da stampare                                 */
/*          color  - Colore della stringa                                  */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void print_system_string (int startx, int starty, char *msg)
{
 unsigned short index,             /* Variabile di comodo */
                x_coord, y_coord;       /* Coordinate effettive */

 /* La stringa  non vuota ? */
 if (msg != NULL)
  /* No, allora scandiamola per tutta la sua lunghezza */
  for (index=0; index<strlen(msg); index++)
   {
    /* Calcolo delle coordinate iniziali di ogni carattere */
    x_coord = startx + index * CHAR_LENGHT;
    y_coord = starty;

    /* Adesso stampiamo il carattere */
    print_system_char (x_coord, y_coord, msg[index]);
   }
}


void main (void)
{
 char *messaggio1="*** Un salutone a tutti ***",
      *messaggio2="***** i lettori di DEV! *****";

 /* Questa variabile la utilizziamo per "centrare" il messaggio */
 unsigned short center;

 /* Inizializiamo la scheda grafica nel modo 320x200x256 */
 Init_Graph_Lib();
 set_vga_mode ();

 /* Prendiamoci il font di sistema */
 get_system_font ();

 /* Stampiamo un messaggio */
 center = (320 - strlen (messaggio1) * X_CHAR_SIZE) / 2;
 print_system_string (center, 80, messaggio1);

 /* Stampiamone un altro, a distanza di sicurezza */
 center = (320 - strlen (messaggio2) * X_CHAR_SIZE) / 2;
 print_system_string (center, 80+Y_CHAR_SIZE, messaggio2);

 /* Attendiamo la pressione di un qualsivoglia tasto */
 getch ();

 /* DOS mon amour... */
 set_text_mode ();
 Close_Graph_Lib();
}
