/*
 *  Programmazione Grafica 3D - Cap. 6
 *  by Andrea Fasano
 *
 * In questo esempio vediamo come "catturare" un font disegnato a mano
 * ed utilizzarlo.
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 */

#include <stdio.h>
#include "demogfx.h"
#include "pcx.h"

/* nota: le dimensione dentro i box sono X_CHAR_SIZE-1 e Y_CHAR_SIZE-1
         in tal caso ogni box racchiude 23x23 pixel.                   */
#define X_CHAR_SIZE 24           /* La dimensione orizzontale del carattere */
#define Y_CHAR_SIZE 24           /* La dimensione verticale del carattere */

#define NUM_CHAR    43           /* Il numero max. dei caratteri */
#define STARTING_ASCII_CODE 48   /* Il primo carattere da noi rappresentato
                                    in questo esempio  il carattere '0', il
                                    quale ha il codice ascii uguale a 48 */
#define TEXT_TRANSPARENT 1       /* 0 == Il carattere cancella completamente
                                         il background;
                                    1 == Il carattere viene copiato in
                                         trasparenza;*/

/* Questo  il nostro nuovo tipo: una matrice per contenere la forma di ogni
   carattere */
typedef unsigned char TYPE_CHAR [X_CHAR_SIZE][Y_CHAR_SIZE];

TYPE_CHAR alfa[NUM_CHAR];                          /* Il nostro "font" */

VIRTUAL       font_pic;          /* L'immagine PCX contenente i caratteri */
TYPE_PALETTE  font_pal[256];     /* La relativa palette */

/*-------------------------------------------------------------------------*/
/* FUNZIONE get_graphic_font                                               */
/* UTILIZZO Riempie un array con caratteri grafici presi da un'immagine    */
/* INPUT    buffer - immagine da cui ritagliare i caratteri                */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void get_graphic_font (VIRTUAL buffer)
{
 int  x, y,             /* Variabili di comodo */
      x1=1, y1=1;       /* Variabili per lo spiazzamento all'interno della
                           immagine. Sono inizializzate ad 1 poich noi non
                           vogliamo anche memorizzare il box bianco che
                           circonda ogni carattere (il quale  appunto spesso
                           1 pixel) */

 char kx=0, ky=0;       /* Variabili per l'incremento della matrice TYPE_CHAR
                           di ogni singolo elemento del nostro array */
 short   index=0;          /* Variabile d'incremento */

 /*
   NB: L'atto di acquisizione dei caratteri grafici presenta diverse
       similitudini con quello effettuato in listato1.c. Le uniche differenze
       di cui dobbiamo tener conto ora sono che non lavoriamo pi su bit ma
       su byte, e che le singole righe che compongono ogni carattere non sono
       pi poste una di seguito all'altra ma sono poste una sotto all'altra
       (in pratica, sono distanziate di 320 byte l'una dall'altra).
 */

 /* Leggiamo il max. numero di caratteri */
 while (index < NUM_CHAR)
  {
   kx = ky = 0;
   /* Memorizziamo il "giusto" pixel */
   for (y=y1; y<(y1+Y_CHAR_SIZE-1); y++)
    {
     for (x=x1; x<(x1+X_CHAR_SIZE-1); x++)
         {
          alfa[index][kx][ky] = buffer[x+y*320];
          kx++;
         }
     ky++;
     kx=0;
    }

   /* Facciamo un controllo per incrementare correttamente sia x1 che y1 */
   x1 += X_CHAR_SIZE;
   /* E' necessario considerare i bordi del box ed il numero di caratteri
      presenti su ogni riga (11 in tal caso) */
   if (x1 > X_CHAR_SIZE*11)
    {
     x1 = 1;
     y1 += Y_CHAR_SIZE;
    }

   index++;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE print_graphic_char                                             */
/* UTILIZZO Stampa un carattere valido sullo schermo                       */
/* INPUT    startx - Ascissa iniziale                                      */
/*          starty - Ordinata iniziale                                     */
/*          ascii_char - Carattere da stampare                             */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void print_graphic_char (int startx, int starty, char ascii_char)
{
 unsigned short x, y,           /* Variabili di comodo */
          x_coord, y_coord;     /* Coordinate effettive */
 char     pixel_color,          /* Colore di ogni pixel del carattere */
          index;                /* Corretto indice dell'array*/

 index = ascii_char - STARTING_ASCII_CODE;

 if (ascii_char >= STARTING_ASCII_CODE &&
     index < NUM_CHAR)   /* Nel caso i caratteri presenti siano meno di 256 */
 for (y=0; y<Y_CHAR_SIZE; y++)
   for (x=0; x<X_CHAR_SIZE; x++)
     {
      x_coord = x + startx;  /* Ascissa reale  */
      y_coord = y + starty;  /* Ordinata reale */

      /* Eseguiamo il clipping */
      if ((x_coord < 320) && (y_coord < 200))
       {
        /* Prendiamo il colore del pixel corrente */
        pixel_color = alfa[index][x][y];

        /* Controlliamo la trasparenza, basta questo OR */
        if ( (pixel_color>0) || (!TEXT_TRANSPARENT) )
           putpixel (x_coord, y_coord, pixel_color);
       }
     }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE print_graphic_string                                           */
/* UTILIZZO Stampa una stringa sullo schermo                               */
/* INPUT    startx - Ascissa iniziale                                      */
/*          starty - Ordinata iniziale                                     */
/*          *msg   - Messaggio da stampare                                 */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void print_graphic_string (int startx, int starty, char *msg)
{
 unsigned short index,                  /* Variabile d'incremento */
                x_coord, y_coord;       /* Coordinate effettive */

 /* La stringa  non vuota ? */
 if (msg != NULL)
  /* No, allora scandiamola per tutta la sua lunghezza */
  for (index=0; index<strlen(msg); index++)
   {
    /* Calcolo delle coordinate iniziali di ogni carattere */
    x_coord = startx + index * X_CHAR_SIZE;
    y_coord = starty;

    /* Adesso stampiamo il carattere */
    print_graphic_char (x_coord, y_coord, msg[index]);
   }
}

void main (void)
{
 /* Qualche messaggio d'esempio. Attenzione! Ricordatevi di usare solo
    le maiuscole! */
 char *messaggio1="UN SALUTONE",
      *messaggio2="A TUTTI",
      *messaggio3="I LETTORI",
      *messaggio4="DI DEV";

 /* Questa variabile la utilizziamo per "centrare" il messaggio */
 unsigned short center;

 /* Carichiamo l'immagine contente i caratteri. Deve essere presente nella
    directory indicata (se no cambiarla) */
 font_pic = load_pcx ("font.pcx", font_pal);
 /* Inizializiamo la scheda grafica nel modo 320x200x256 */
 Init_Graph_Lib();
 set_vga_mode ();

 /* Settiamo la palette */
 set_palette (font_pal);

 /* Prendiamoci il font da noi disegnato */
 get_graphic_font (font_pic);

 /* Stampiamo un messaggio */
 center = (320 - strlen (messaggio1) * X_CHAR_SIZE) / 2;
 print_graphic_string (center, 60, messaggio1);

 /* Stampiamone il secondo, a distanza di sicurezza */
 center = (320 - strlen (messaggio2) * X_CHAR_SIZE) / 2;
 print_graphic_string (center, 60+Y_CHAR_SIZE, messaggio2);

 /* Ora il terzo... */
 center = (320 - strlen (messaggio3) * X_CHAR_SIZE) / 2;
 print_graphic_string (center, 60+Y_CHAR_SIZE*2, messaggio3);

 /* Infine l'ultimo! */
 center = (320 - strlen (messaggio4) * X_CHAR_SIZE) / 2;
 print_graphic_string (center, 60+Y_CHAR_SIZE*3, messaggio4);

 /* Attendiamo la pressione di un qualsivoglia tasto */
 getch();

 /* Indietro al DOS */
 set_text_mode ();
 Close_Graph_Lib();
}
