/*
 *  Programmazione Grafica 3D - Cap. 8
 *  by Andrea Fasano
 *
 * Libreria per la gestione dell'alta risoluzione secondo lo standard
 * grafico VESA VBE 2.0.
 *
 *  Rispetto a cap. 7 sono state aggiunte alcune primitive grafiche
 *  (anche per il page flipping), specifiche per per la risoluzione 640x480.
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 */

#ifndef _DEMOVBE_
#define _DEMOVBE_

/* SELETTORE DEL TIPO DI ACCESSO ALLA MEMORIA VIDEO */
#define SECURE_ACCESS   0     /* CAMBIARE CON 0 SOLO NELLA VERSIONE FINALE */

#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <dpmi.h>
#include <go32.h>

#if     !SECURE_ACCESS
#include <sys/nearptr.h>
#else
#include <sys/farptr.h>
#endif

/* La direttiva `__attribute__((packed))' serve per impedire a GCC di
   inserire dei gaps tra alcuni membri di uno struct per allinearli
   opportunamente alla scopo di velocizzarne l'accesso. Per maggiori
   informazioni, vedi le faq di djgpp
                 "How to move structs returned by real-mode services?"
                 e altri (cerca)                                       */
/* questa define serve solo a scopo di brevit                         */
#define pack __attribute__((packed))

/* macro di conversione di un puntatore reale ad un
   puntatore per il modo protetto (protected mode pointer) */
#define real2pm_ptr(realptr, pmptr) pmptr = (realptr&0xffff)+((realptr&0xffff0000)>>12);

/****************************************************************************/
/*  DEFINIZIONE DEI TIPI                                                    */
/****************************************************************************/
/*
  Tipo VbeInfoBlock: in esso sono contenute tutte le caratteristiche del
                     display grafico a nostra disposizione.
*/
typedef struct {
   unsigned char      VbeSignature[4]       pack; // Deve essere 'VBE2'
   unsigned short int VbeVersion            pack; // Versione del driver
   char* OemName                            pack; // Nome della scheda grafica
   unsigned char      Capabilities[4]       pack; // Caratt. display grafico
   unsigned short int*SupportedModes        pack; // Puntatore alla lista
                                                  // dei modi supportati
   unsigned short int TotalMemory           pack; // Memoria a bordo della
                                                  // scheda
   unsigned short int OemSoftwareRev        pack; // Livello revisione VBE
   char* OemVendorName                      pack; // Nome del produttore
   char* OemProductName                     pack; // Nome del prodotto
   char* OemProductRev                      pack; // Livello revisione display

   unsigned char      reserved[222]         pack; // Riservato
   unsigned char      OemData[256]          pack; //
} VbeInfoBlock;

/*
  Tipo ModeInfoBlock: in esso sono contenute tutte le caratteristiche
                      del modo grafico che vogliamo attivare.
*/
typedef struct {
  unsigned short int ModeAttributes    pack; // Specifiche del modo
  unsigned char      WinAAttributes    pack; // Caratt. della window A
  unsigned char      WinBAttributes    pack; // Caratt. della window B
  unsigned short int WinGranularity    pack; // Granularit -> window
  unsigned short int WinSize           pack; // Dimensione  -> window
  unsigned short int WinASegment       pack; // Indirizzo window A
  unsigned short int WinBSegment       pack; // Indirizzo window B
  void	      (*WPF) (signed long int page) pack; // Indirizzo funzione
  unsigned short int BytesPerScanLine   pack;

  unsigned short int XResolution       pack; // Larghezza in pixel
  unsigned short int YResolution       pack; // Altezza in pixel
  unsigned char      XCharSize         pack; // Larghezza carattere
  unsigned char      YCharSize         pack; // Altezza carattere
  unsigned char      NumberOfPlanes    pack; // Numero dei planes disponibili
  unsigned char      BitsPerPixel      pack; // Num. bit per ogni pixel
  unsigned char      NumberOfBanks     pack; // Num. dei banchi presenti
  unsigned char      MemoryModel       pack; // Tipo di memoria utilizzato
  unsigned char      BankSize          pack; // Dimensione di ogni banco
  unsigned char      NumberOfImagePages pack;// Num. -1 di schermate
  unsigned char      Reserved          pack; // Riservato

  unsigned char      RedMaskSize       pack; // Maschera per rosso
  unsigned char      RedFieldPosition  pack; // Posizione bit rosso
  unsigned char      GreenMaskSize     pack; // Maschera per verde
  unsigned char      GreenFieldPosition pack;// Posizione bit verde
  unsigned char      BlueMaskSize      pack; // Maschera per blu
  unsigned char      BlueFieldPosition pack; // Posizione bit blu
  unsigned char      RsvdMaskSize      pack; //
  unsigned char      RsvdFieldPosition pack; //
  unsigned char      DirectColorModeInfo pack; // Caratt. colori modo diretto

  unsigned long int  PhysBasePtr       pack; // Linear Frame Buffer
  unsigned long int  OffScreenMemoryOffset pack;// Offset mem. "fuori schermo"
  unsigned short int OffScreenMemSize  pack; // Mem. disponibile   ""  ""
  unsigned char      Reserved2 [206]   pack; // Riservato
} ModeInfoBlock;

/****************************************************************************/
/*  VARIABILI GLOBALI                                                       */
/****************************************************************************/
/* Nota: queste informazioni sono settate rispettivamente da get_vbe_info()
   e da get_mode_info. La prima basta chiamarla una sola volta. La seconda
   va chiamata ogni volta che cambiate modo grafico, PRIMA di set_vbe_mode()
   Potete accedere NORMALMENTE a tutti i campi di VbeInfo e ModeInfo       */
VbeInfoBlock  VbeInfo;                 // Informazioni sul VBE 2.0
ModeInfoBlock ModeInfo;                // Informazioni sul modo

#if     SECURE_ACCESS
short int Vid_Selector;
unsigned int baseaddress;               // Indirizzo della prima pagina
#else
unsigned char      *vbe_screen;        // Indirizzo iniziale dello schermo
unsigned char      *videobase;         // Indirizzo della prima pagina
#endif

_go32_dpmi_registers regs;

/* Numeri e tipo delle modalit grafiche */
char VideoModeNames [96][40] = {
   "100 - 640x400      256 colors", "101 - 640x480      256 colors",
   "102 - 800x600       16 colors", "103 - 800x600      256 colors",
   "104 - 1024x768      16 colors", "105 - 1024x768     256 colors",
   "106 - 1280x1024     16 colors", "107 - 1280x1024    256 colors",
   "108 - 80x60         Text Mode", "109 - 132x25        Text Mode",
   "10A - 132x43        Text Mode", "10B - 132x50        Text Mode",
   "10C - 132x60        Text Mode", "10D - 320x200   32,768 colors",
   "10E - 320x200   65,536 colors", "10F - 320x200      16m colors",
   "110 - 640x480   32,768 colors", "111 - 640x480   65,536 colors",
   "112 - 640x480      16m colors", "113 - 800x600   32,768 colors",
   "114 - 800x600   65,536 colors", "115 - 800x600      16m colors",
   "116 - 1024x768  32,768 colors", "117 - 1024x768  65,536 colors",
   "118 - 1024x768     16m colors", "119 - 1280x1024 32,768 colors",
   "11A - 1280x1024 65,536 colors", "11B - 1280x1024    16m colors",
   "11C - 640x350      256 colors", "11D - 640x350   32,768 colors",
   "11E - 640x400   32,768 colors", "11F - 640x350   65,536 colors",
   "120 - 640x400   65,536 colors", "121 - 640x350      16m colors",
   "122 - 640x400      16m colors", "123 - 1600x1200     16 colors",
   "124 - 1600x1200    256 colors", "125 - 1600x1200 32,768 colors",
   "126 - 1600x1200 65,536 colors", "127 - 1600x1200    16m colors",
   "128 - 640x480   16m colors(*)", "129 - 800x600   16m colors(*)",
   "12A - 1024x768  16m colors(*)", "12B - 1280x1024 16m colors(*)",
   "12C - 1600x1200 16m colors(*)", "12D - 320x240   32,768 colors",
   "12E - 320x400   32,768 colors", "12F - 360x200   32,768 colors",
   "130 - 360x240   32,768 colors", "131 - 360x400   32,768 colors",
   "132 - 320x240   65,536 colors", "133 - 320x400   65,536 colors",
   "134 - 360x200   65,536 colors", "135 - 360x240   65,536 colors",
   "136 - 360x400   65,536 colors", "137 - 320x240      16m colors",
   "138 - 320x400      16m colors", "139 - name n/a",
   "13A - name n/a",                "13B - name n/a",
   "13C - name n/a",                "13D - name n/a",
   "13E - name n/a",                "13F - name n/a",
   "140 - name n/a",                "141 - name n/a",
   "142 - 640x350   16m colors(*)", "143 - 640x400   16m colors(*)",
   "144 - name n/a",                "145 - name n/a",
   "146 - name n/a",                "147 - name n/a",
   "148 - name n/a",                "149 - name n/a",
   "14A - name n/a",                "14B - name n/a",
   "14C - name n/a",                "14D - name n/a",
   "14E - name n/a",                "14F - name n/a",
   "150 - 640x350       16 colors", "151 - 640x400       16 colors",
   "152 - 640x480       16 colors", "153 - 320x200      256 colors",
   "154 - 320x240      256 colors", "155 - 320x400      256 colors",
   "156 - 360x200      256 colors", "157 - 360x240      256 colors",
   "158 - 360x400      256 colors", "159 - name n/a",
   "15A - name n/a",                "15B - name n/a",
   "15C - name n/a",                "15D - name n/a",
   "15E - name n/a",                 "15F - name n/a"
};

/****************************************************************************/
/*  PROTOTIPI                                                               */
/****************************************************************************/
/* Ottiene le informazioni sul driver presente, e le memorizza in VbeInfo */
void get_vbe_info (void);

/* Stampa le informazioni contenute in VbeInfo in modo formattato */
void show_vbe_info (void);

/* Ottiene le informazioni sul modo Vbe_Mode e le memorizza in ModeInfo */
void get_mode_info (unsigned short int Vbe_Mode);

/* Stampa le informazioni contenute in ModeInfo */
void show_mode_info (unsigned short int Vbe_Mode);

/* Attiva il modo Vbe_Mode e il Linear Frame Buffer */
void set_vbe_mode (unsigned short int Vbe_Mode);

/* Setta la lunghezza dell'intera memoria grafica */
void set_logical_scanline (int length);

/* Setta il punto di partenza nella visualizzazione dell'intera memoria */
void set_display_start (int pixel, int line);

/* Copia di schermo sul video */
void copy_screen_to_vbe (char *source);

/* Copia di schermo in schermo */
void vbe_flip_screen (char *destination, char *source);

/* Cancella la pagina attiva */
void blank_screen (void);

/* Cancella tutte le pagine e rende attiva e visibile la pagina 0 */
void blank_all (void);

/* Stampa di pixel */
void vbe_put_pixel (int x, int y, char color);

/* Setta la pagina attiva */
void set_active_page (int page_number);

/* Setta la pagina visiva */
void set_visual_page (int page_number);

/* Controllo d'errore su chiamata funzione VBE (funzione ad uso "interno") */
void vbe_check_status (unsigned short vbe_function_result);

/****************************************************************************/
/*  FUNZIONI                                                                */
/****************************************************************************/
/* estrae una stringa puntata da un puntatore per il modo protetto,
   e allocata in memoria convenzionale, la copia nello heap del prg.
   e ne restituisce il puntatore - usata da get_vbe_info() */
char *extract_str(char* pmcptr)
{
  char str[200];    // 200 dovrebbe bastare!
  char ch;          // un carattere
  short i;          // contatore caratteri
  char *cptr;       // puntatore di comodo

  i=0;
  cptr=pmcptr;

  if (cptr==NULL)
    return NULL;

  do
  {
    _dosmemgetb((int)cptr++,1, &ch); // Sposta un byte dalla memoria
    str[i++] = ch;                   // convenzionale ad un array di
                                     // caratteri nella memoria protetta
  }while ((ch != '\0')&&(i < 200-1));
  str[i]='\0'; // termina la stringa, se non  stata terminata (i=200)

  return strdup(str);
}

/* per copiare la tabella dei modi supportati dalla memoria convenzionale
   nello heap; come la funzione precedente  usata da get_vbe_info()
   e non deve essere chiamata dall'utente.                              */
short *extract_modetable(void)
{
 short int mode_table[97];
 short int mode, i;
 short int *ptr;

 i = 0;
 ptr = VbeInfo.SupportedModes; // Ottiene il puntatore che punta ad una lista
                               // di tutti i numeri dei modi video vesa
                               // supportati e poi ricrea questa lista nella
                               // memoria dinamica direttamente indirizzabile
                               // dal programma

/*
 SupportedMode  un puntatore ad una lista di numeri di modi video vesa
 Alla fine della lista vi  uno short int che vale -1.
*/

// Copiamo ogni numero di modo video nello spazio di indirizzamento
// del nostro programma.

 do
 {                                  // Copia ogni numero
  _dosmemgetw((int)ptr++,1, &mode); // Sposta una word dalla memoria conv. ad
  mode_table[i++] = mode;           // un array di short int nella memoria protetta.
 }while ((mode != -1)&&(i < 96));
 mode_table[i]=-1;          // terminatore
 /*

  Adesso mode_table[] contiene tutti i numeri dei modivideo vesa.
  Non ci resta che copiarla nello heap e restituire il puntatore.
 */
 i=2*(i+1); // i=numero di byte della tabella (compreso il terminatore)
 ptr=(short int*)malloc(i); // alloca sufficiente memoria
 /* copia dall'array nello heap */
 memcpy(ptr,mode_table,i);
 return ptr;
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE get_vbe_info                                                   */
/* UTILIZZO Ottiene le informazioni sul display grafico.                   */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void get_vbe_info (void)
{
 unsigned int seg, ofs;

 VbeInfo.VbeSignature[0] = '2'; // Per prima cosa inizializziamo la struttura
 VbeInfo.VbeSignature[1] = 'E'; // da notare che la rappresentazione
 VbeInfo.VbeSignature[2] = 'B'; // little-endian
 VbeInfo.VbeSignature[3] = 'V';

 VbeInfo.VbeVersion = 0x200;    // Anche questa inizializzazione ci vuole.
                                // altrimenti la struttura VbeInfo non sar
                                // correttamente riempita dall'interrupt VBE
                                // chiamato

 /*

  Poich le funzioni VBE operano in modalit reale, dobbiamo copiare
  la struttura inizializzata prima (VbeInfo) nella memoria convenzionale
  prima della chiamata, in modo che la struttura pu essere riempita dalle
  funzioni VESA che operano in real mode.

 */

 dosmemput(&VbeInfo, sizeof(VbeInfo), __tb);

 /* vedi djgppfaq: 18.2 How to use buffers with DOS/BIOS services */

 seg = __tb;     // Calcola l'indirizzo real mode del buffer.
 ofs = __tb;
 ofs &= 0xffff;
 seg &= 0xffff0000;
 seg >>= 4;

 /* in DJGPP  facile simulare la chiamata all'int reale:
    basta chiamare una funzione DPMI fatta apposta       */
 regs.d.eax = 0x4f00;
 regs.x.es = seg;
 regs.x.di = ofs;
 __dpmi_simulate_real_mode_interrupt(0x10, &regs); // Ottiene le vesa info.


 dosmemget(__tb, sizeof(VbeInfo),&VbeInfo); // Spostiamo la struttura di nuovo
                                            // nello spazio di indirizzamento
                                            // del nostro programma

 vbe_check_status (regs.d.eax & 0xffff);

 if(VbeInfo.VbeVersion < 0x200) // Il Vesa driver non  VBE 2.0 compatibile,
 {                              // perci esci.
  printf("VESA implementation is not compatible with VBE 2.0.\n");
  exit(-1);
 }


 /*

  La struttura VbeInfo viene riempita con valori di puntatori che sono validi
  solo in modalit reale, cio tutti i puntatori sono puntatori a 16 bit
  (16 bit per il segmento e 16 bit per l'offset). Quindi dobbiamo convertirli
  in puntatori per il modo protetto. Quest'ultimi non possono essere usati
  come normali puntatori. Dobbiamo usare un selettore di segmento con un
  valore di base pari a zero e un limite di almeno 1MB. Questo tipo di
  selettore di segmento  gi disponibile. Lo possiamo ottenere semplicemente
  chiamando questa funzione della libreria C di DJGPP (a cui rimando):
  _go32_conventional_mem_selector().

 */

 /* usiamo la macro definita all'inizio del codice
    per effettuare queste conversioni ripetitive  */
 real2pm_ptr((int)VbeInfo.OemName,(int)VbeInfo.OemName);
 real2pm_ptr((int)VbeInfo.SupportedModes, (int)VbeInfo.SupportedModes);
 real2pm_ptr((int)VbeInfo.OemVendorName, (int)VbeInfo.OemVendorName);
 real2pm_ptr((int)VbeInfo.OemProductName,(int)VbeInfo.OemProductName);
 real2pm_ptr((int)VbeInfo.OemProductRev,(int)VbeInfo.OemProductRev);

 /* di modo che l'utente non deve preoccuparsi di usare selettori
    e pu accedere normalmente anche ai campi puntatori della
    struttura VbeInfo, facciamo una copia nello heap del programma
    delle stringhe e dell'array di short che contiene i codici
    dei modi video supportati e cambiamo di nuovo i campi
    puntatori della struttura VbeInfo */

 VbeInfo.OemName = extract_str(VbeInfo.OemName);
 VbeInfo.OemVendorName = extract_str(VbeInfo.OemVendorName);
 VbeInfo.OemProductName = extract_str(VbeInfo.OemProductName);
 VbeInfo.OemProductRev = extract_str(VbeInfo.OemProductRev);

 /* anche la lista dei modi disponibili viene copiata nello heap */
 VbeInfo.SupportedModes=extract_modetable();
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE show_vbe_info                                                  */
/* UTILIZZO Stampa a video alcune caratteristiche del display grafico.     */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void show_vbe_info (void)
{
 unsigned short int  list,
          *VideoModePtr;            /** Variabili di comodo **/
          char modec;

 printf ("\n");
 printf ("-VESA VBE 2.0 General Info by Andrea Fasano ------------\n");

 printf ("Software ................. %s\n", VbeInfo.OemName);

 if (VbeInfo.OemVendorName == NULL) printf ("Vendor Name .............. n/a\n");
 else printf ("Vendor Name .............. %s\n", VbeInfo.OemVendorName);

 if (VbeInfo.OemProductName == NULL) printf ("Product Name ............. n/a\n");
 else printf ("Product Name ............. %s\n", VbeInfo.OemProductName);

 if (VbeInfo.OemProductRev == NULL) printf ("Product Revision ......... n/a\n");
 else printf ("Product Revision ......... %s\n", VbeInfo.OemProductRev);

 printf ("--------------------------------------------------------\n");

 printf ("Driver Version............ %d.%d\n", VbeInfo.VbeVersion/100,
                                               VbeInfo.VbeVersion%100);

 printf ("Total Memory OnBoard...... %d Kb\n", VbeInfo.TotalMemory<<6);

 printf ("Capabilities:\n");
 (VbeInfo.Capabilities[0]&1) ? printf ("   - DAC is fixed width, with 6 bits per primary color\n")
                           : printf ("   - DAC width is switchable to 8 bits per primary color\n");
 (VbeInfo.Capabilities[0]&2) ? printf ("   - Controller is VGA compatible\n")
                           : printf ("   - Controller is not VGA compatible\n");
 (VbeInfo.Capabilities[0]&4) ? printf ("   - Normal RAMDAC operation\n")
                           : printf ("   - RAMDAC recommends programming during blank period only\n");
 /* La lista dei modi supportati */
 /* da notare che l'espressione C *++VideoModePtr incrementa il
    puntatore ad unsigned short VideoModePtr (incrementa l'indirizzo
    di 2) e poi lo dereferenzia (quindi  lo stesso di *(++VideoModePtr) */
 list=0;
 VideoModePtr=VbeInfo.SupportedModes;
 printf ("\nList of modes supported:\n");
 do
  {
   modec = (unsigned char) *VideoModePtr;

   if (modec<96)
   {
      printf ("    %s   ", VideoModeNames[modec]);/*ignora eventuale warning*/
      list++;
      if (list % 2 == 0) printf ("\n");
   }
 }  while (*++VideoModePtr != 0xffff);

 printf ("\nTotal modes supported: %d\n", list);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE get_mode_info                                                  */
/* UTILIZZO Ottiene le informazioni sul modo grafico desiderato ed ottiene */
/*          l'indirizzo del linear frame buffer.                           */
/* INPUT    unsigned short int Vbe_Mode - N. modalit grafica              */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void get_mode_info(unsigned short int Vbe_Mode)
{
 int ofs, seg;

 seg = __tb;   // __tb  un buffer che si trova nella "memoria dos". Questo
 ofs = __tb;   //  necessario perch il driver vesa non pu accedere alla
               // memoria al di sopra di 1 MB (e il nostro programma 
               //  di solito allocato al di sopra di 1MB).

 ofs &= 0xffff;     // Crea un puntatore reale (segment:offset = 16:16)
 seg &= 0xffff0000; // dall'indirizzo del buffer.
 seg >>= 4;

 regs.d.eax = 0x4f01;   // Funzione per ottenere le informazioni su un modo
                        // modo video vesa specificato in una struttura
                        // (ModeInfoBlock) che sar utilizzata da altre
 regs.d.ecx = Vbe_Mode; // funzioni.
 regs.x.es = seg;
 regs.x.di = ofs;
 __dpmi_simulate_real_mode_interrupt(0x10, &regs);

 dosmemget(__tb,256,&ModeInfo); // Le informazioni si trovano nella memoria
                                // DOS perci devono essere spostate nello
                                // spazio di indirizzamento del nostro
                                // programma.

/*
   Ogni servizio degli interrupt VBE ritorna un codice dopo che  stato
   chiamato. Se la chiamata ha avuto successo, il codice di ritorno  0x4f.
   La funzione vbe_check_status() termina l'esecuzione segnalando il tipo
   di errore incontrato, se c' stato un errore, altrimenti non fa niente.
*/

 // Il codice di ritorno  contenuto in EAX
 vbe_check_status(regs.d.eax & 0xffff);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE show_mode_info                                                 */
/* UTILIZZO Stampa a video alcune caratteristiche del modo desiderato.     */
/* INPUT    unsigned short int Vbe_Mode - N. modalit grafica              */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void show_mode_info (unsigned short int Vbe_Mode)
{
 printf ("\n\n DEV VESA VBE 2.0 Driver Mode Info by Andrea Fasano, Marzo 1997");
 printf ("\n Modo: %dx%dx%d (%d bits)", ModeInfo.XResolution,
                                        ModeInfo.YResolution,
                                        ModeInfo.BitsPerPixel<<5,
                                        ModeInfo.BitsPerPixel);

 (ModeInfo.ModeAttributes&1)? printf ("\n   - Mode SUPPORTED in hardware configuration")
                             : printf ("\n   - Mode NOT supported in hardware configuration");
 (ModeInfo.ModeAttributes&8)? printf ("\n   - Color Mode")
                             : printf("\n   - Monochrome mode");
 (ModeInfo.ModeAttributes&16)? printf ("\n   - Graphic Mode")
                             : printf ("\n   - Text Mode");
 (ModeInfo.ModeAttributes&32)? printf ("\n   - Not VGA compatible")
                             : printf ("\n   - VGA compatible");
 (ModeInfo.ModeAttributes&64)? printf ("\n   - Windowed memory mode available")
                             : printf ("\n   - Windowed memory mode NOT available");
 (ModeInfo.ModeAttributes&128)? printf ("\n   - Linear Frame Buffer available :-)")
                             : printf ("\n   - Linear Frame Buffer NOT available :-(");

 printf ("\n\n Bytes Per ScanLine %d", ModeInfo.BytesPerScanLine);

 printf ("\n\n Window A Attributes:");
 (ModeInfo.WinAAttributes&1)? printf ("\n   - Window is relocatable")
                             : printf ("\n   - Window is not relocatable");
 (ModeInfo.WinAAttributes&2)? printf ("\n   - Window is readable")
                             : printf ("\n   - Window is not readable");
 (ModeInfo.WinAAttributes&4)? printf ("\n   - Window is writeable")
                             : printf ("\n   - Window is not writeable");
 printf ("\n\n Window B Attributes:");
 (ModeInfo.WinBAttributes&1)? printf ("\n   - Window is relocatable")
                             : printf ("\n   - Window is not relocatable");
 (ModeInfo.WinBAttributes&2)? printf ("\n   - Window is readable")
                             : printf ("\n   - Window is not readable");
 (ModeInfo.WinBAttributes&4)? printf ("\n   - Window is writeable")
                             : printf ("\n   - Window is not writeable");

 printf ("\n\n Window granularity %d Kb", ModeInfo.WinGranularity);
 printf ("\n Window size        %d Kb", ModeInfo.WinSize);

 printf ("\n\n Memory Model is ");
 switch (ModeInfo.MemoryModel)
  {
   case 0x00: printf ("Text Mode"); break;
   case 0x01: printf ("CGA Graphics"); break;
   case 0x02: printf ("Hercules Graphics"); break;
   case 0x03: printf ("Planar"); break;
   case 0x04: printf ("Packed Pixel"); break;
   case 0x05: printf ("Non-chain 4, 256 color"); break;
   case 0x06: printf ("Direct Color"); break;
   case 0x07: printf ("YUV Color"); break;
   default: printf ("to be defined by OEM");
  }

 printf ("\n Number of Banks %d", ModeInfo.NumberOfBanks);
 printf ("\n Bank Size       %d Kb\n", ModeInfo.BankSize);

 printf ("\n Red   Mask Size %d (%d)", ModeInfo.RedMaskSize, ModeInfo.RedFieldPosition);
 printf ("\n Green Mask Size %d (%d)", ModeInfo.GreenMaskSize, ModeInfo.GreenFieldPosition);
 printf ("\n Blue  Mask Size %d (%d)", ModeInfo.BlueMaskSize, ModeInfo.BlueFieldPosition);

 printf ("\n\nColor Mode Info:");
 (ModeInfo.DirectColorModeInfo&1)? printf ("\n   - Color ramp is programmable")
                                  : printf ("\n   - Color ramp is fixed");
 (ModeInfo.DirectColorModeInfo&2)? printf ("\n   - Bits in Rsvd are usable")
                                  : printf ("\n   - Bits in Rsvd are reserved");
 putchar('\n');
}

/* specifica per DJGPP, usata da set_vbe_mode() */
void init_vesa(void)
{
 __dpmi_meminfo mem_info;

#if     SECURE_ACCESS
 /*
   Questa forse  la parte pi difficile di tutte da capire. Questa funzione
   alloca un descrittore, mappa la memoria video, imposta il valore di base
   del selettore all'indirizzo fisico della memoria mappata. Imposta i
   diritti di accesso del selettore, in modo che la memoria possa essere
   usata sia in lettura che in scrittura.

   Se vuoi fare delle funzioni grafiche come plottare un pixel, devi
   sempre usare Vid_Selector per leggere o scrivere nella memoria video.
   un esempio di una putpixel in linguaggio assembly potrebbe essere:

   putpixel(int x, int y, char color)

    mov  eax, [y]
    mul  dword ptr [xpix]    ; Xpix  la risoluzione orizzontale.
    add  eax, [x]
    mov  bl, [color]
    mov  gs, [vid_sel]

    mov  [gs:eax], bl


   Ho scelto questo approccio di allocare un nuovo selettore per la memoria
   video, perch penso che sia il pi sicuro, ma non  l'unico modo di
   usare un linear frame buffer. Possiamo anche mappare il linear frame
   buffer in modo che pu essere acceduto direttamente tramite DS,
   effettivamente disabilitando tutta la sicurezza propria della
   programmazione in modo protetto. In tal caso il nostro segmento dati
   avrebbe un limite di segmento di 4GB, perci un qualsiasi puntatore
   a cui  assegnato un indirizzo sbagliato potrebbe fare seri
   danneggiamenti.
   Una funzione grafica errata pu far andare in crash il sistema.
   Invece, usando uno speciale descrittore, possiamo scrivere solo
   nella memoria video e non possiamo in alcun modo sovrascrivere il
   codice del programma o i suoi dati (cose che possono accadere
   piuttosto facilmente con i puntatori "selvaggi").

 */


 // Mappa il linear frame buffer nella memoria fisica.

 mem_info.address = ModeInfo.PhysBasePtr; // L'inizio del linear frame buffer
 mem_info.size = (VbeInfo.TotalMemory << 16); // Numero di bytes di memoria
                                              // video (numero di blocchi da
                                              // 64k * 65536).

 if(-1==__dpmi_physical_address_mapping(&mem_info))
 {
  printf("Error mapping physical memory.\n");
  exit(-1);
 }

 // Crea un nuovo descrittore per il linear frame buffer.

 if(1==(Vid_Selector = __dpmi_allocate_ldt_descriptors(1)))
  {
   printf("Cannot allocate descriptor for video memory.\n");
   exit(-1);
  }

 // Imposta i diritti di accesso del segmento del nuovo descrittore.

 if(-1==__dpmi_set_descriptor_access_rights(Vid_Selector, 0x40f3))
 {
  printf("Cannot set segment access rights.\n");
  exit(-1);
 }

 // salviamo l'indirizzo di base (della prima pagina):
 // ci serve per poter effettuare il page flipping
 baseaddress = mem_info.address;

 // Si spiega da s.

 if(-1==__dpmi_set_segment_base_address(Vid_Selector, baseaddress))
  {
   printf("Cannot set segment base.\n");
   exit(-1);
  }

 // Si spiega da s.

 if(-1==(__dpmi_set_segment_limit(Vid_Selector, (VbeInfo.TotalMemory<<16)|0xfff ) ))
 {
  printf("Cannot set segment limit.\n");
  exit(-1);
 }

 /* settiamo questo selettore una volta per tutte, in modo da non
    doverlo passare ogni volta alle funzioni _far*               */
 _farsetsel(Vid_Selector);

#else /* il metodo trucchettoso super-veloce ! */
 /* dapprima bisogna mappare, come col metodo precedente */

 // Mappa il linear frame buffer nella memoria fisica.

 mem_info.address = ModeInfo.PhysBasePtr; // L'inizio del linear frame buffer
 mem_info.size = (VbeInfo.TotalMemory << 16); // Numero di bytes di memoria
                                              // video (numero di blocchi da
                                              // 64k * 65536).

 if(-1==__dpmi_physical_address_mapping(&mem_info))
 {
  printf("Error mapping physical memory.\n");
  exit(-1);
 }

     if (__djgpp_nearptr_enable())
        vbe_screen = (unsigned char *)(__djgpp_conventional_base + mem_info.address);
     else
     {
        printf("video memory direct access not permetted by this DPMI server\n"
               "try to recompile with SECURE_ACCESS=1\n");
        exit(-1);
     }

 videobase=vbe_screen;  /* vbe_screen punta all'inizio della pagina 0 */
#endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_vbe_mode                                                   */
/* UTILIZZO Attiva una modalit grafica VBE con il linear frame buffer.    */
/* INPUT    unsigned short int Vbe_Mode - N. modalit grafica              */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_vbe_mode (unsigned short int Vbe_Mode)
{
 /* Set Vesa Vbe mode */
 /* settiamo il modo e creiamo un descrittore apposta */

 init_vesa();

 regs.d.eax = 0x4f02;
 /* Attiviamo il Linear Frame Buffer */
 regs.d.ebx = Vbe_Mode | 0x4000;  // Facciamo OR con 0x4000 perch vogliamo
                                  // la versione con linear frame buffer del
                                  // modo e non quella banked.

 __dpmi_simulate_real_mode_interrupt(0x10, &regs);

 vbe_check_status (regs.d.eax & 0xffff);      // controllo e gestione errore

 /* Inizializziamo il sistema, ponendo la pagina grafica attiva e visibile
     coincidenti con la pagina 0 */
 set_logical_scanline (640);

 /* facciamo pulizia su tutte le pagine disponibili,
    in modo da non avere "spazzatura" e settiamo la
    prima pagina (numero 0) come attiva e visibile */
 blank_all();
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_logical_scanline                                           */
/* UTILIZZO Definisce la lunghezza in pixel dell'intera memoria grafica    */
/*          non deve essere inferiore a quella della risoluzione attiva!   */
/*          (data da ModeInfo.XResolution)                                 */
/* INPUT    lenght - Lunghezza della scanline                              */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_logical_scanline (int length)
{
  union REGS registers;

  /* 0x06 = Set Scanline Lenght */
  registers.w.ax  = 0x4f06;
  /* Definiamo la lunghezza in pixel */
  registers.h.bl = 0x00;
  /* Immagazziniamo la lunghezza */
  registers.w.cx = length;

  int386 (0x10, &registers, &registers);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_display_start                                              */
/* UTILIZZO Definisce il punto iniziale di visualizzazione dello schermo   */
/*          "sopra" l'intera memoria grafica.                              */
/* INPUT    int start_pixel - pixel iniziale                               */
/*          int start_line  - linea iniziale                               */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_display_start (int start_pixel, int start_line)
{
  union REGS registers;

  /* 0x07 = Set Display Start */
  registers.w.ax = 0x4f07;
  /* bh  riservato e deve essere 0 */
  registers.h.bh = 0x00;
  /* Vogliamo settare la partenza */
  registers.h.bl = 0x00;

  /* Pixel iniziale */
  registers.w.cx = start_pixel;
  /* Linea iniziale */
  registers.w.dx = start_line;

  int386 (0x10, &registers, &registers);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE blank_screen                                                   */
/* UTILIZZO Cancella la pagina correntemente attiva                        */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void blank_screen (void)
{
  #if SECURE_ACCESS
     int i;
     for (i=0; i<640*480/4; i++)
        _farnspokel(i<<2, 0);
  #else
     int i;
     for (i=0; i<640*480/4; i++)
        ((int*)vbe_screen)[i]=0; /* azzera 4 byte per volta con accesso diretto */
  #endif
}

void blank_all (void)
{
 int i;

 /* per tutte le pagine disponibili in questa modalit e per questa scheda */
 for (i=0; i<ModeInfo.NumberOfImagePages+1; i++)
 {
  set_active_page(i);   /* attiva la pagina */
  blank_screen();       /* clear it! */
 }

 /* imposta la prima pagina come attiva e visibile */
 set_visual_page (0);
 set_active_page (0);
}

void vbe_put_pixel (int x, int y, char color)
{
  #if SECURE_ACCESS
     _farnspokeb(x + ((y<<9)+(y<<7)), color);
  #else
     vbe_screen[x+ ((y<<9)+(y<<7))]=color;

     // conversione in assembly
     // cos (conversione diretta dal codice assembly inline
     // per il Watcom (vedi libro pag. 121) non ne vale la pena,
     // non  molto pi veloce con le ottimizzazioni attivate
     //  ( praticamente lo stesso come efficienza del codice
     //  generato dal compilatore, persino con le ottimizzazioni
     //  disattivate). Significa che DJGPP  meglio di Watcom?
     //  Io penso di s (comunque per esserne certo dovrei
     //  avere Watcom e confrontarli; ma non voglio spendere soldi
     //  per un compilatore commerciale visto che c' DJGPP free!)

//     asm("movl  %1, %%edx
//          movl  _vbe_screen, %%edi
//          shll  $9, %1
//          shll  $7, %%edx
//          addl  %1, %%edi
//          addl  %%edx, %%edi
//          addl  %0, %%edi
//          stosb"
//     : /* no outputs */
//     : /* inputs */
//       "c" (x), "b" (y), "a" (color)
//     : /* globbered registers */
//       "edx", "edi");
  #endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE copy_screen_to_vbe                                             */
/* UTILIZZO Trasferisce il contenuto della generica area di memoria        */
/*          sul monitor                                                    */
/* INPUT    source - Indirizzo iniziale dello schermo virtuale             */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void copy_screen_to_vbe (char* source)
 {
  #if SECURE_ACCESS
     movedata(_my_ds(), (int)source, Vid_Selector, 0, 640*480);
  #else
     int i;
     for (i=0; i<640*480/4; i++)
        ((int*)vbe_screen)[i]=((int*)source)[i];
  #endif
 }

/*-------------------------------------------------------------------------*/
/* FUNZIONE vbe_flip_screen                                                */
/* UTILIZZO Copia il contenuto dello schermo virtuale source nello schermo */
/*          dest                                                           */
/* INPUT    source - Schermo sorgente                                      */
/*          dest   - Schermo destinatario                                  */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void vbe_flip_screen (char *destination, char *source)
{
   memcpy(destination,source,307200);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_active_page                                                */
/* UTILIZZO Seleziona la pagina attiva                                     */
/*          attenzione: questa funzione, per ragioni di efficienza, NON    */
/*          controlla che il numero di pagina specificato sia valido,      */
/*          perci  vostra responsabilit. A seconda del modo grafico e   */
/*          della memoria della scheda, questo numero varia.               */
/*          E' dato da:                                                    */
/*                              ModeInfo.NumberOfImagePages+1              */
/* INPUT    page_number - numero della pagina                              */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_active_page (int page_number)
{
#if     SECURE_ACCESS
 /* cambiamo la base del selettore Vid_Selector */
 if(-1==__dpmi_set_segment_base_address(Vid_Selector,
                                        baseaddress+640*480 * page_number))
  {
   set_text_mode();
   printf("Cannot set segment base.\n");
   exit(-1);
  }

 /* rendiamo predefinito questo nuovo selettore */
 /* non  necessario!! */
// _far setsel(Vid_Selector);

#else
 vbe_screen = videobase + 640*480 * page_number;
#endif
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE set_visual_page                                                */
/* UTILIZZO Seleziona la pagina visiva                                     */
/* INPUT    page_number - numero della pagina                              */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void set_visual_page (int page_number)
{
  union REGS registers;

  /* 0x07 = Set Display Start */
  registers.w.ax = 0x4f07;
  /* bh  riservato e deve essere 0 */
  registers.h.bh = 0x00;
  /* Vogliamo settare la partenza */
  registers.h.bl = 0x00;

  /* Pixel iniziale */
  registers.w.cx = 0;
  /* Linea iniziale */
  registers.w.dx = 480 * page_number;

  int386 (0x10, &registers, &registers);
}

void vbe_check_status (unsigned short vbe_function_result)
{
 if (vbe_function_result != 0x004f)
  {
   set_text_mode ();
   printf ("\nAttenzione! Si  verificato un errore: (%x) ", vbe_function_result);
   switch (vbe_function_result)
    {
     case 0x014f : printf ("generico."); break;
     case 0x024f : printf ("incompatibilit hardware."); break;
     case 0x034f : printf ("funzione non valida nel modo corrente."); break;
     default     : printf ("errore di tipo sconosciuto.");
    }
   exit (0);
  }
}

#endif
