/*
 *  Programmazione Grafica 3D - Cap. 8
 *  by Andrea Fasano
 *
 * Tale esempio  puramente *dimostrativo*, in quanto vuole mostrare
 * esclusivamente un esempio di utilizzo della libreria "demovbe2.h"
 * NB: Le risoluzioni contemplate sono tutte a 256 colori
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 */

#include <stdio.h>
#include "demogfx.h"
#include "demovbe2.h"

char FIRE_HEIGHT = 1;          /* Step per la sfumatura verso il nero*/

unsigned char *fb, MAX_HEIGHT;
int X_SIZE, Y_SIZE, SCREEN_SIZE;

/*char fb[Y_SIZE+2][X_SIZE];            Il buffer atto a contenere il fuoco. Esso 
                                 2 linee in pi rispetto alla normale
                                 grandezza dello schermo, in quanto non verranno
                                 mostrate dal momento che sono un po' bruttine */
TYPE_PALETTE firepal[256];     /* La palette del fuoco */


/*-------------------------------------------------------------------------*/
/* FUNZIONE generate_fire_palette                                          */
/* UTILIZZO Setta la palette firepal[] in modo appropriato per "simulare"  */
/*          i colori della fiamma                                          */
/* INPUT    pal - palette da riempire                                      */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void generate_fire_palette (TYPE_PALETTE *pal)
{
 int n;

 /* I primi 64 colori sono organizzati secondo una sfumatura graduale dal
    nero al rosso vivo (il primo colore sar nero) */
 for (n=0; n<64; n++)
  {
   pal[n].r = n;
   pal[n].g = 0;
   pal[n].b = 0;
  }

 /* Altri 64 colori li utilizziamo per le sfumature di giallo della fiamma */
 for (n=64; n<128; n++)
  {
   pal[n].r = 63;
   pal[n].g = n-64;
   pal[n].b = 0;
  }

 /* altri 64 colori li utilizziamo per realizzare
    una sfumatura dal giallo al bianco vivo */
 for (n=128; n<192; n++)
  {
   pal[n].r = 63;
   pal[n].g = 63;
   pal[n].b = n-128;
  }

 /* gli ultimi 64 colori a nostra disponizione vengono utilizzati
    solo per le posizioni pi basse del fuoco, apparendo sotto
    forma di macchioline, e perci ci mettiamo tutti dei bianchi. */
 for (n=192; n<256; n++)
  {
   pal[n].r = 63;
   pal[n].g = 63;
   pal[n].b = 63;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE move_fire                                                      */
/* UTILIZZO Crea il fuoco dal buffer fb e lo "muove" verso l'alto          */
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void move_fire (void)
{
 int x;

 for (x=0; x<SCREEN_SIZE; x++)
  {
   fb[x] = (fb[x+1] + fb[x+X_SIZE-1] + fb[x+X_SIZE] + fb[x+X_SIZE+1])>>2;

   if (fb[x] > MAX_HEIGHT) fb[x] -= FIRE_HEIGHT;
   else if (fb[x] > 1) fb[x] = 0;
  }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE update_last_lines                                              */
/* UTILIZZO Aggiorna le ultime 2 linee del buffer fb[][] con valori casuali*/
/* INPUT    Nessuno                                                        */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void update_last_lines (void)
{
 int x;

 for (x=0; x<X_SIZE; x++)
   {
    fb[SCREEN_SIZE+x] = rand () % 256;
    fb[SCREEN_SIZE+X_SIZE+x] = rand () % 256;
   }
}

void main (int argc, char *argv[])
{
 int vbe_mode, tmp;

 /* Check */
 if (argv[1] == NULL)
  {
   printf ("\nUtilizzo: fuoco [mode number]");
   printf ("\nmode number ::== 0  - 320x200");
   printf ("\n                 1  - 320x240");
   printf ("\n                 2  - 320x400");
   printf ("\n                 3  - 360x200");
   printf ("\n                 4  - 360x240");
   printf ("\n                 5  - 360x400");
   printf ("\n                 6  - 640x350");
   printf ("\n                 7  - 640x400");
   printf ("\n                 8  - 640x480");
   printf ("\n                 9  - 800x600");
   printf ("\n                 10 - 1024x768");
   printf ("\n                 11 - 1280x1024");
   printf ("\n                 12 - 1600x1200");
   exit (0);
  }
 else
  {
   tmp = atoi (argv[1]);
   switch (tmp)
    {
     case 0: vbe_mode = 0x153; break;
     case 1: vbe_mode = 0x154; break;
     case 2: vbe_mode = 0x155; break;
     case 3: vbe_mode = 0x156; break;
     case 4: vbe_mode = 0x157; break;
     case 5: vbe_mode = 0x158; break;
     case 6: vbe_mode = 0x11c; break;
     case 7: vbe_mode = 0x100; break;
     case 8: vbe_mode = 0x101; break;
     case 9: vbe_mode = 0x103; break;
     case 10: vbe_mode = 0x105; break;
     case 11: vbe_mode = 0x107; break;
     case 12: vbe_mode = 0x124; break;
     default: vbe_mode = 0x153;
    }
  }

 /* Attivazione risoluzione */
 get_vbe_info ();
 get_mode_info (vbe_mode);
 set_vbe_mode (vbe_mode);

 /* Settaggio dei parametri */
 Y_SIZE = scr_info.y_res;
 X_SIZE = scr_info.x_res;

 SCREEN_SIZE = X_SIZE * Y_SIZE;
 MAX_HEIGHT = Y_SIZE / 100;

 /* Creiamo il buffer del fuoco */
 fb = create_screen (X_SIZE, Y_SIZE+2);
 memset (fb, 0, scr_info.size);

 generate_fire_palette (firepal);   /* Creiamo un'opportuna palette */
 set_palette (firepal);             /* Settiamola */

 while (!kbhit())                   /* Attendiamo la pressione di un tasto */
  {
   update_last_lines ();
   move_fire ();                    /* Ciclo principale del fuoco */
   copy_screen_to_vbe (fb, scr_info.size);
  }


 free (fb);                         /* Liberiamo memoria */
 set_text_mode ();                  /* Oops! Indietro al buon vecchio DOS.. */
}
