#include <go32.h>
#include <dpmi.h>
#define real2pm_ptr(realptr, pmptr) pmptr = (realptr&0xffff)+((realptr&0xffff0000)>>12);


// Structures used by vesa functions.

typedef struct
{
 char vbesignature[4]     __attribute__((packed));
 short int vbeversion     __attribute__((packed));
 char* oemstringptr        __attribute__((packed));
 int  capabilities        __attribute__((packed));
 short int *videomodeptr  __attribute__((packed));
 short int totalmemory    __attribute__((packed));
 short int oemsoftwarerev __attribute__((packed));
 char* oemvendornameptr    __attribute__((packed));
 char*  oemproductnameptr   __attribute__((packed));
 char*  oemproductrevptr    __attribute__((packed));
 char reserved[222]       __attribute__((packed));
 char oemdata[256]        __attribute__((packed));

}vbeinfoblock;

typedef struct
{
 short int modeattributes __attribute__((packed)); //     dw ?      ; mode attributes
 char winaattributes __attribute__((packed)); //     db ?      ; window A attributes
 char winbattributes __attribute__((packed));      //db ?      ; window b attributes
 short int wingranularity __attribute__((packed));      //dw ?      ; window granularity
 short int winsize __attribute__((packed));             //dw ?      ; window size
 short int winasegment __attribute__((packed));         //dw ?      ; window A start segment
 short int winbsegment __attribute__((packed));         //dw ?      ; window b start segment
 int  winfuncptr __attribute__((packed));          //dd ?      ; pointer to window function
 short int bytesperscanline __attribute__((packed));    //dw ?      ; bytes per scan line

     //; mandatory information for vbe 1.2 and above
 short int xresolution __attribute__((packed));         //dw ?      ; horizontal resolution in pixels or chars
 short int yresolution __attribute__((packed));         //dw ?      ; vertical resolution in pixels or chars
 char xcharsize __attribute__((packed));           //db ?      ; character cell width in pixels
 char ycharsize __attribute__((packed));           //db ?      ; character cell height in pixels
 char numberofplanes __attribute__((packed));      //db ?      ; number of memory planes
 char bitsperpixel __attribute__((packed));        //db ?      ; bits per pixel
 char numberofbanks __attribute__((packed));       //db ?      ; number of banks
 char memorymodel __attribute__((packed));         //db ?      ; memory model type
 char banksize __attribute__((packed));            //db ?      ; bank size in Kb
 char numberofImagepages __attribute__((packed));  //db ?      ; number of images
 char reserved1 __attribute__((packed));            //db ?      ; reserved for
     //; direct color fields (required for direct/6 and YUv/7 memory models)
 char redmasksize __attribute__((packed));         //db ?      ; size of direct color red mask in bits
 char redfieldposition __attribute__((packed));    //db ?      ; bit position of lsb of red mask
 char greenmasksize __attribute__((packed));       //db ?      ; size of direct color green mask in bits
 char greenfieldposition __attribute__((packed));  //db ?      ; bit position of lsb of green mask
 char bluemasksize __attribute__((packed));        //db ?      ; size of direct color blue mask in bits
 char bluefieldposition __attribute__((packed));   //db ?      ; bit position of lsb of blue mask
 char rsvdmasksize __attribute__((packed));        //db ?      ; size of direct color reserved mask in bits
 char rsvdfieldposition __attribute__((packed));   //db ?      ; bit position of lsb of reserved mask
 char directcolormodeInfo __attribute__((packed)); //db ?      ; direct color mode attributes

 //    ; mandatory information for vbe 2.0 and above
 int  physbaseptr __attribute__((packed));       //  dd ?      ; physical address for flat frame buffer
 int  offscreenmemoffset __attribute__((packed));  //dd ?      ; pointer to start of off screen memory
 short int offscreenmemsize __attribute__((packed));//    dw ?      ; amount of off screen memory in 1k units
 char reserved2[206] __attribute__((packed));        //            db 206 dup (?)  ; remainder of modeInfoblock
}vbe_modeinfo;





int videobase, bytes_per_screen;
short int vid_selector;

_go32_dpmi_registers regs;


vbeinfoblock vbeinfo;
vbe_modeinfo modeinfo;


int get_mode_info(int mode)
{
 int ofs, seg;

 seg = __tb;   // __tb is a buffer located in "dos memory". This buffer
 ofs = __tb;   // is used because the vesa driver cannot acces memory
               // above 1 MB (your program is most likely to be located above 1MB).

 ofs &= 0xffff;     // Make a real pointer (segment:offse = 16:16)
 seg &= 0xffff0000; // from the address of the buffer.
 seg >>= 4;

 regs.d.eax = 0x4f01; // Get the modeinfo of a certain vesa video mode,
                      // this is a structure which contains.
 regs.d.ecx = mode;   // information needed by other functions below.
 regs.x.es = seg;
 regs.x.di = ofs;
 __dpmi_simulate_real_mode_interrupt(0x10, &regs);

 dosmemget(__tb,256,&modeinfo); // This info is located in dos memory, so
                                // it has to be moved to your program's
                                // addres space.


/*
   Every VBE interrupt service returns a code after it is called. If the call
   succeeded, the return code is 0x4f. If the call succeeded, this function
   (get_mode_info) returns 0, on failure it returns -1.
*/

 if(0x4f != (regs.d.eax & 0xffff)) // The return code is contained by EAX
  return(-1);
 else
  return(0);

}

int get_vbe_info(void)
{
 unsigned int seg, ofs;

 vbeinfo.vbesignature[0] = '2'; // First off initialize the structure.
 vbeinfo.vbesignature[1] = 'E';
 vbeinfo.vbesignature[2] = 'B';
 vbeinfo.vbesignature[3] = 'V';

 vbeinfo.vbeversion = 0x200;    // Same here. Must be done, else the structure
                                // vbeinfo structure will not be filled
                                // correctly.


 /*

  Because VBE funtions operate in real mode, we first have to move the
  initialized structure to real-mode address space, so the structure can
  be filled by the vesa function in real mode.

 */

 dosmemput(&vbeinfo, sizeof(vbeinfo), __tb);


 seg = __tb;     // Calculate real mode addres of the buffer.
 ofs = __tb;
 ofs &= 0xffff;
 seg &= 0xffff0000;
 seg >>= 4;

 regs.d.eax = 0x4f00;
 regs.x.es = seg;
 regs.x.di = ofs;
 __dpmi_simulate_real_mode_interrupt(0x10, &regs); // Get vesa info.



 dosmemget(__tb, sizeof(vbeinfo),&vbeinfo); // Move the structure back to
                                            // your program's address space.

 if(vbeinfo.vbeversion < 0x200) // Vesa driver is not VBE 2.0 compatible, so exit.
 {
  printf("VESA implementation is not compatible with VBE 2.0.\n");
  exit(-1);
 }


 /*

  The VBE info structure is filled with values which are only valid in
  real mode, so all pointers are 16 bit pointers (16 bits segment and 16 bit ofset).
  We first have to convert them to protected mode pointers. These new pointers
  may not be used as normal pointers. You have to use a segment selector with
  a segment base value of zero, and a limit of at least 1MB. This segment selector
  is already  available. You can get it by calling :
  _go32_conventional_mem_selector().

 */

 real2pm_ptr((int)vbeinfo.oemstringptr,(int)vbeinfo.oemstringptr);
 real2pm_ptr((int)vbeinfo.videomodeptr, (int)vbeinfo.videomodeptr);
 real2pm_ptr((int)vbeinfo.oemvendornameptr, (int)vbeinfo.oemvendornameptr);
 real2pm_ptr((int)vbeinfo.oemproductnameptr,(int)vbeinfo.oemproductnameptr);
 real2pm_ptr((int)vbeinfo.oemproductrevptr,(int)vbeinfo.oemproductrevptr);

 return(regs.d.eax & 0xffff);
}


void init_globals(void)
{

 // Just initialize some stuff. Not really needed, but i find it handy.
 // Yeah, a really waste of memory to use globals, eh ? (NOT)

 videobase = modeinfo.physbaseptr;
 bytes_per_screen = modeinfo.xresolution*modeinfo.yresolution;
}


int init_vesa(void)
{
 int i;
 __dpmi_meminfo mem_info;


 /*
   This is for many people the hardest part of all. This function allocates
   a descriptor, maps the video memory , sets the base value of the selector
   to the physical addres of this mapped memory. Sets the acces rights of
   the selector, so that the memory can be used to read and write to it.

   If you want to make graphical functions such as putpixel, you
   must always use the vid_selector for reading or writing to video memory.
   an example of an assembly language putpixel could be :


   putpixel(int x, int y, char color)

    mov  eax, [y]
    mul  dword ptr [xpix]    ; Xpix is the horizontal resolution.
    add  eax, [x]
    mov  bl, [color]
    mov  gs, [vid_sel]

    mov  [gs:eax], bl




   I've choosen for this aproach of allocating a new selector for the video
   memory, because I think it's the safest approach. It is not the only
   way to use a linear frame buffer. You can also map the linear frame
   buffer, so that it can be accessed by DS, effectively disabling all
   the security things that are essential to protected mode programming.
   Your data segment would have a segment limit of 4GB, so any pointer
   with a bad address assigned to it could do some serious damage.
   And especially badly written graphical functions can screw things up
   very bad. By using a special descriptor, you can only write to video
   memory, and not to your program or your data (things that can happen
   very easily with "wild" pointers).

 */


 // Map the linear frame buffer to physical memory.


 mem_info.address = videobase;  // The beginning of the linear frame buffer
 mem_info.size = (vbeinfo.totalmemory << 16); // Number of bytes of video
                                              // memory (number of 64k block * 65536).

 if(-1==__dpmi_physical_address_mapping(&mem_info))
 {
  printf("Error mapping physical memory.\n");
  exit(-1);
 }
 videobase = mem_info.address;


 // Creata a new descriptor for the linear frame buffer.

 if(1==(vid_selector = __dpmi_allocate_ldt_descriptors(1)))
  {
   printf("Cannot allocate descriptor for video memory.\n");
   exit(-1);
  }

 // Set the segment access rights of the new descriptor.

 if(-1==__dpmi_set_descriptor_access_rights(vid_selector, 0x40f3))
 {
  printf("Cannot set segment acces rights.\n");
  exit(-1);
 }

 // Explains itself.

 if(-1==__dpmi_set_segment_base_address(vid_selector, videobase))
  {
   printf("Cannot set segment base.\n");
   exit(-1);
  }

 // Explains itself.

 if(-1==(__dpmi_set_segment_limit(vid_selector, (vbeinfo.totalmemory<<16)|0xfff ) ))
 {
  printf("Cannot set segment limit.\n");
  exit(-1);
 }


 videobase = 0;

 return(0);
}

int setvesamode(int xres, int yres, int bpp)
{
 short int mode_table[200];
 short int mode, i;
 short int *ptr;


 if(0x4f != get_vbe_info()) // First get the general VESA info.
 {
  return(-1);
 }

 i = 0;
 ptr = vbeinfo.videomodeptr; // Get the pointer that points to a list of all
                             // videomode numbers and then search this list
                             // for the requested vesa mode.
/*

 The videomodeptr is a pointer to a list of
 vesa videomode numbers. At the end of the list is a short int with a value
 of -1.

*/

// Move every videomode number to your program's address space.

 do
 {                                  // Move every
  _dosmemgetw((int)ptr++,1, &mode); // Move a word from conv. memory to
  mode_table[i++] = mode;           // an array of short int's in PM memory.
 }while ((mode != -1)&&(i <= 198));

 /*

  Now mode_table[] contains all vesa videomode numbers. Now we have to
  check all these video modes, if there is one available with the
  dimensions, and the right color depth passed to this function.

 */


 i = 0;

 while((unsigned short int)mode_table[i] != 0xffff)
 {

  mode = mode_table[i++];

  if (get_mode_info(mode) == 0)


/*
     Is this the mode which we're looking for ?

     If so, set the mode and create a descriptor for it.
     If it isn't the mode we're looking for, skip it and check the next
     mode.
*/


    if((modeinfo.xresolution == xres)&&(modeinfo.yresolution == yres)&&
      (modeinfo.bitsperpixel == bpp))

     {
      regs.d.eax = 0x4f02;
      regs.d.ebx = mode | 0x4000; // We OR with 0x4000 because we need
                                  // the linear frame buffer version of
                                  // the mode and not banked version.

      __dpmi_simulate_real_mode_interrupt(0x10, &regs);

      init_globals();
      init_vesa();
      return(0);
     }
 }

 // We searched through all available modes but didn't find the correct mode

 printf("\nMode not availeble, not able to set videomode.\n");
 exit(-1);
}
